import random
import sys
import os
import errno
import shutil
import copy
import traceback
import threading
import asyncio
import subprocess
import aiofiles
import aiosqlite
import json
import psutil
import datetime
import time
import tempfile
import signal
from enum import Enum
from . import logging
from .nethelpers import get_addr_to, get_default_addr, get_localhost, address_to_ip_port
from .net_classes import WorkerResources
# from .worker_task_protocol import WorkerTaskServerProtocol
from .exceptions import NotEnoughResources, ProcessInitializationError, WorkerNotAvailable, AlreadyRunning,\
    InvocationMessageWrongInvocationId, InvocationMessageAddresseeTimeout, InvocationMessageError
# from .scheduler_task_protocol import SchedulerTaskClient
from .worker_messsage_processor import WorkerMessageProcessor
from .scheduler_message_processor import SchedulerWorkerControlClient
from .worker_invocation_protocol import WorkerInvocationProtocolHandlerV10, WorkerInvocationServerProtocol
#from .worker_pool_protocol import WorkerPoolClient
from .worker_pool_message_processor import WorkerPoolControlClient
from .broadcasting import await_broadcast
from .invocationjob import InvocationJob
from .config import get_config, create_default_user_config_file
from . import environment_resolver
from .enums import WorkerType, WorkerState, ProcessPriorityAdjustment
from .paths import config_path
from .local_notifier import LocalMessageExchanger
from .filelock import FileCoupledLock
from .process_utils import create_process, kill_process_tree
from . import db_misc
from .misc import DummyLock
from .net_messages.address import AddressChain, DirectAddress
# from .net_messages.impl.tcp_message_processor import TcpMessageProcessor
from .net_messages.exceptions import MessageTransferError
from .defaults import worker_pool_port as default_worker_pool_port, \
                      scheduler_port as default_scheduler_port, \
                      worker_start_port as default_worker_start_port

from .worker_runtime_pythonpath import lifeblood_connection
import inspect

from typing import Any, Optional, Dict, Tuple, Union


is_posix = not sys.platform.startswith('win')


async def create_worker(scheduler_address: AddressChain, *,
                        child_priority_adjustment: ProcessPriorityAdjustment = ProcessPriorityAdjustment.NO_CHANGE,
                        worker_type: WorkerType = WorkerType.STANDARD,
                        singleshot: bool = False,
                        worker_id: Optional[int] = None,
                        pool_address: Optional[Tuple[str, int]] = None):
    worker = Worker(scheduler_address, child_priority_adjustment=child_priority_adjustment, worker_type=worker_type, singleshot=singleshot, worker_id=worker_id, pool_address=pool_address)

    await worker.start()  # note that server is already started at this point
    return worker


class Worker:
    def __init__(self, scheduler_addr: AddressChain, *,
                 child_priority_adjustment: ProcessPriorityAdjustment = ProcessPriorityAdjustment.NO_CHANGE,
                 worker_type: WorkerType = WorkerType.STANDARD,
                 singleshot: bool = False,
                 scheduler_ping_interval: float = 10,
                 scheduler_ping_miss_threshold: int = 6,
                 worker_id: Optional[int] = None,
                 pool_address: Optional[AddressChain] = None):
        """

        :param scheduler_addr:
        :param scheduler_port:
        :param worker_type:
        :param singleshot:
        """
        config = get_config('worker')
        self.__logger = logging.get_logger('worker')
        self.log_root_path: str = ''
        for self.log_root_path in (os.path.expandvars(config.get_option_noasync('worker.logpath', config_path('logs', 'worker'))),
                                   os.path.join(tempfile.gettempdir(), 'lifeblood', 'worker_logs')):
            logs_ok = True
            try:
                if not os.path.exists(self.log_root_path):
                    os.makedirs(self.log_root_path, exist_ok=True)
            except PermissionError:
                logs_ok = False
            except OSError as e:
                if e.errno == errno.EACCES:
                    logs_ok = False
            logs_ok = logs_ok and os.access(self.log_root_path, os.W_OK)
            if logs_ok:
                break
            self.__logger.warning(f'could not use location {self.log_root_path} for logs, trying another...')
        else:
            raise RuntimeError('could not initialize logs directory')
        self.__logger.info(f'using {self.log_root_path} for logs')

        self.__status = {}
        self.__scheduler_db_uid: int = 0  # unsigned 64bit int
        self.__running_process: Optional[asyncio.subprocess.Process] = None
        self.__running_process_start_time: float = 0
        self.__running_task: Optional[InvocationJob] = None
        self.__running_task_progress: Optional[float] = None
        self.__running_awaiter = None
        self.__previous_notrunning_awaiter = None  # here we will temporarily save running_awaiter before it is set to None again when task canceled or finished, to avoid task being GCd while in work
        self.__message_processor: Optional[WorkerMessageProcessor] = None
        self.__local_invocation_server: Optional[asyncio.Server] = None
        self.__local_invocation_server_address_string: str = ''

        self.__local_shared_dir = config.get_option_noasync("local_shared_dir_path", os.path.join(tempfile.gettempdir(), 'lifeblood_worker', 'shared'))
        self.__resource_db_lock = FileCoupledLock('worker_resources', self.__local_shared_dir)
        self.__resource_db_path = os.path.join(self.__local_shared_dir, 'resources.db')
        self.__my_resources = WorkerResources(cpu_count=config.get_option_noasync('resources.cpu_count'),
                                              cpu_mem=config.get_option_noasync('resources.cpu_mem'),
                                              gpu_count=config.get_option_noasync('resources.gpu_count'),
                                              gpu_mem=config.get_option_noasync('resources.gpu_mem'))
        self.__task_changing_state_lock = asyncio.Lock()
        self.__stop_lock = threading.Lock()
        self.__start_lock = asyncio.Lock()  # cant use threading lock in async methods - it can yeild out, and deadlock on itself
        self.__where_to_report: Optional[AddressChain] = None
        self.__ping_interval = scheduler_ping_interval
        self.__ping_missed_threshold = scheduler_ping_miss_threshold
        self.__ping_missed = 0
        self.__scheduler_addr = scheduler_addr
        self.__scheduler_pinger = None
        self.__components_stop_event = asyncio.Event()
        self.__extra_files_base_dir = None
        self.__my_addr: Optional[AddressChain] = None
        self.__worker_id = worker_id
        self.__pool_address: Optional[AddressChain] = pool_address
        if self.__worker_id is None and self.__pool_address is not None \
                or self.__worker_id is not None and self.__pool_address is None:
            raise RuntimeError('pool_address must be given together with worker_id')

        self.__worker_task_comm_queues: Dict[str, asyncio.Queue] = {}

        self.__worker_type: WorkerType = worker_type
        self.__singleshot: bool = singleshot or worker_type == WorkerType.SCHEDULER_HELPER

        self.__child_priority_adjustment = child_priority_adjustment
        # this below is a placeholder solution. the easiest way to implement priority lowering without testing on different platrofms
        if self.__child_priority_adjustment == ProcessPriorityAdjustment.LOWER:
            if sys.platform.startswith('win'):
                assert hasattr(psutil, 'BELOW_NORMAL_PRIORITY_CLASS')
                psutil.Process().nice(psutil.BELOW_NORMAL_PRIORITY_CLASS)
            else:
                psutil.Process().nice(10)

        # deploy a copy of runtime module somewhere in temp
        rtmodule_code = inspect.getsource(lifeblood_connection)

        filepath = os.path.join(tempfile.gettempdir(), 'lifeblood', 'lifeblood_runtime', 'lifeblood_connection.py')
        os.makedirs(os.path.dirname(filepath), exist_ok=True)
        existing_code = None
        if os.path.exists(filepath):
            with open(filepath, 'r') as f:
                existing_code = f.read()

        if existing_code != rtmodule_code:
            with open(filepath, 'w') as f:
                f.write(rtmodule_code)
        self.__rt_module_dir = os.path.dirname(filepath)

        self.__stopping_waiters = []
        self.__finished = asyncio.Event()
        self.__started = False
        self.__started_event = asyncio.Event()
        self.__stopped = False

    def message_processor(self) -> WorkerMessageProcessor:
        return self.__message_processor

    def scheduler_message_address(self) -> AddressChain:
        return self.__scheduler_addr

    async def start(self):
        if self.__started:
            return
        if self.__stopped:
            raise RuntimeError('already stopped, cannot start again')

        async with self.__start_lock:
            abort_start = False

            # start local server for invocation api connections
            loop = asyncio.get_event_loop()
            localhost = get_localhost()
            localport_start = 10101
            localport_end = 11111
            localport = None
            for _ in range(localport_end - localport_start):  # big but finite
                localport = random.randint(localport_start, localport_end)
                try:
                    self.__local_invocation_server = await loop.create_server(
                        lambda: WorkerInvocationServerProtocol(self, [WorkerInvocationProtocolHandlerV10(self)]),
                        localhost,
                        localport
                    )
                    break
                except OSError as e:
                    if e.errno != errno.EADDRINUSE:
                        raise
                    continue
            else:
                raise RuntimeError('could not find an opened port!')
            self.__local_invocation_server_address_string = f'{localhost}:{localport}'

            # start message processor
            my_ip = get_addr_to(self.__scheduler_addr.split_address()[0])
            my_port = default_worker_start_port()
            for i in range(1024):  # big but finite
                try:
                    self.__message_processor = WorkerMessageProcessor(self, (my_ip, my_port))
                    await self.__message_processor.start()
                    break
                except OSError as e:
                    if e.errno != errno.EADDRINUSE:
                        raise
                    my_port += 1
                    continue
            else:
                raise RuntimeError('could not find an opened port!')

            if self.__pool_address:
                self.__my_addr = AddressChain.join_address((self.__pool_address, DirectAddress.from_host_port(my_ip, my_port)))
            else:
                self.__my_addr = AddressChain.from_host_port(my_ip, my_port)

            # now report our address to the scheduler
            try:
                with SchedulerWorkerControlClient.get_scheduler_control_client(self.__scheduler_addr, self.__message_processor) as client:  # type: SchedulerWorkerControlClient
                    self.__scheduler_db_uid = await client.say_hello(self.__my_addr, self.__worker_type, self.__my_resources)
            except MessageTransferError as e:
                self.__logger.error('error connecting to scheduler during start')
                abort_start = True
            #
            # and report to the pool
            try:
                if self.__worker_id is not None:
                    assert self.__pool_address is not None
                    with WorkerPoolControlClient.get_worker_pool_control_client(self.__pool_address, self.__message_processor) as wpclient:  # type: WorkerPoolControlClient
                        await wpclient.report_state(self.__worker_id, WorkerState.IDLE)
            except ConnectionError as e:
                self.__logger.error('error connecting to worker pool during start')
                abort_start = True

            self.__scheduler_pinger = asyncio.create_task(self.scheduler_pinger())
            self.__started = True
            self.__started_event.set()
        if abort_start:
            self.__logger.error('error during stating worker, aborting!')
            self.stop()
        else:
            self.__logger.info('worker started')

    def is_started(self):
        return self.__started

    def wait_till_starts(self):  # we can await this function cuz it returns a future...
        return self.__started_event.wait()

    def stop(self):
        async def _send_byebye():
            try:
                self.__logger.debug('saying bye to scheduler')
                with SchedulerWorkerControlClient.get_scheduler_control_client(self.__scheduler_addr, self.__message_processor) as client:  # type: SchedulerWorkerControlClient
                    await client.say_bye(self.__my_addr)
            except MessageTransferError:  # if scheduler or route is down
                self.__logger.info('couldn\'t say bye to scheduler as it seem to be down')
            except Exception:
                self.__logger.exception('couldn\'t say bye to scheduler for unknown reason')

        if not self.__started or self.__stopped:
            return
        with self.__stop_lock:  # NOTE: there is literally no threading in worker, so this is excessive
            self.__logger.info('STOPPING WORKER')
            self.__components_stop_event.set()

            async def _finalizer():
                await self.__scheduler_pinger  # to ensure pinger stops and won't try to contact scheduler any more
                await self.cancel_task()  # then we cancel task, here we still can report it to the scheduler.
                # no new tasks will be picked up cuz __stopped is already set
                self.__local_invocation_server.close()
                await _send_byebye()  # saying bye, don't bother us. (some delayed comms may still come through to the __server
                await self.__local_invocation_server.wait_closed()
                self.__message_processor.stop()
                await self.__message_processor.wait_till_stops()
                self.__logger.info('message processor stopped')

            self.__stopping_waiters.append(asyncio.create_task(_finalizer()))
            self.__finished.set()
            self.__stopped = True

    async def wait_till_stops(self):
        # if self.__scheduler_pinger is not None:
        #     #try:
        #     await self.__scheduler_pinger
        #     #except asyncio.CancelledError:
        #     #    self.__logger.debug('wait_to_finished: scheduler_pinger was cancelled')
        #     #    #raise
        #     self.__scheduler_pinger = None
        # await self.__server.wait_closed()
        await self.__finished.wait()
        self.__logger.info('server closed')
        await self.__scheduler_pinger
        self.__logger.info('pinger closed')
        # await self.__running_resource_checker
        # self.__logger.info('resource checker closed')
        for waiter in self.__stopping_waiters:
            await waiter

    def get_log_filepath(self, level, invocation_id: int = None):  # TODO: think of a better, more generator-style way of returning logs
        if self.__running_task is None and invocation_id is None:
            return os.path.join(self.log_root_path, f'db_{self.__scheduler_db_uid:016x}', 'common', level)
        else:
            return os.path.join(self.log_root_path, f'db_{self.__scheduler_db_uid:016x}', 'invocations', str(invocation_id or self.__running_task.invocation_id()), level)

    async def delete_logs(self, invocation_id: int):
        path = os.path.join(self.log_root_path, f'db_{self.__scheduler_db_uid:016x}', 'invocations', str(invocation_id or self.__running_task.invocation_id()))
        await asyncio.get_event_loop().run_in_executor(None, shutil.rmtree, path)  # assume that deletion MAY take time, so allow util tasks to be processed while we wait

    # This number of methods are related to message exchange mechanism between local workers
    # it was used to exchange resource usage information, but is not used anymore
    # however should it ever be needed - it can be uncommented and adjusted according to needs
    #
    # async def _start_resource_checker(self):
    #     """
    #     starts resource checker task, ensures everything is ready and returns task to await for
    #     """
    #     os.makedirs(self.__local_shared_dir, exist_ok=True)
    #     async with self.__resource_db_lock:
    #         async with aiosqlite.connect(self.__resource_db_path, timeout=30) as con:
    #             con.row_factory = aiosqlite.Row
    #             await con.executescript(db_misc.worker_resource_db_init_script)
    #     await self.check_update_resources()
    #     self.__local_negotiator = LocalMessageExchanger('worker_resources', self.__local_message_exchanger_callback)
    #     self.__local_negotiator.start()
    #     return asyncio.create_task(self.__resource_checker_wait_till_stops())
    #
    # async def __resource_checker_wait_till_stops(self):
    #     """
    #     this task should be running as part of worker.start
    #     it checks and negociates resources with other worker processes on the same machine
    #     """
    #     timeout = await get_config('worker').get_option('resources_recheck_interval', 90)
    #     stop_task = asyncio.create_task(self.__components_stop_event.wait())
    #     while not self.__components_stop_event.is_set():
    #         done, _ = await asyncio.wait((stop_task,), timeout=timeout, return_when=asyncio.FIRST_COMPLETED)
    #         if stop_task in done:  # so we're done
    #             break
    #         # otherwise - timeout happened, so just check resources
    #         await self.check_update_resources()
    #     self.__local_negotiator.stop()
    #     negotiator = self.__local_negotiator
    #     self.__local_negotiator = None
    #     await negotiator.wait_till_stopped()
    #
    # async def __local_message_exchanger_callback(self, message):
    #     return await self.check_update_resources()
    #
    # async def check_update_resources(self, *, already_locked=False):
    #     self.__logger.debug('checking available resources...')
    #     my_pid = os.getpid()
    #     res = WorkerResources()
    #     resource_lock = DummyLock() if already_locked else self.__resource_db_lock
    #     async with resource_lock:
    #         self.__logger.debug('got lock' if not already_locked else 'already locked')
    #         if self.__running_task is not None:
    #             res -= self.__running_task.requirements().to_min_worker_resources()
    #         async with aiosqlite.connect(self.__resource_db_path, timeout=30) as con:
    #             con.row_factory = aiosqlite.Row
    #             to_delete = []
    #             async with con.execute('SELECT * FROM resources WHERE pid != ?', (my_pid,)) as cur:
    #                 async for row in cur:
    #                     if not psutil.pid_exists(row['pid']):
    #                         to_delete.append(row['pid'])
    #                         continue
    #                     res.cpu_count -= row['cpu_count']
    #                     res.cpu_mem -= row['cpu_mem']
    #                     res.gpu_count -= row['gpu_count']
    #                     res.gpu_mem -= row['gpu_mem']
    #
    #             await con.executemany('DELETE FROM resources WHERE pid == ?', [(pid,) for pid in to_delete])
    #             await con.commit()
    #
    #         if self.__my_resources == res:
    #             self.__logger.debug('no need to inform scheduler about resources')
    #         else:
    #             self.__my_resources = res
    #             await self._inform_scheduler_about_resources()
    #
    # async def _inform_scheduler_about_resources(self):
    #     assert self.__my_resources.is_valid()
    #     # not updating scheduler any more.
    #     # scheduler keeps track of resouces by itself
    #
    # async def _set_local_resource_usage(self, resouces: WorkerResources, *, already_locked=False):
    #     resource_lock = DummyLock() if already_locked else self.__resource_db_lock
    #     async with resource_lock:
    #         async with aiosqlite.connect(self.__resource_db_path, timeout=30) as con:
    #             con.row_factory = aiosqlite.Row
    #             await con.execute('INSERT OR REPLACE INTO resources(pid, cpu_count, cpu_mem, gpu_count, gpu_mem) VALUES (?, ?, ?, ?, ?)',
    #                               (os.getpid(),
    #                                resouces.cpu_count,
    #                                resouces.cpu_mem,
    #                                resouces.gpu_count,
    #                                resouces.gpu_mem))
    #             await con.commit()
    #     if self.__local_negotiator:
    #         await self.__local_negotiator.send_sync_event()
    #
    # async def _unset_local_resource_usage(self, *, already_locked=False):
    #     resource_lock = DummyLock() if already_locked else self.__resource_db_lock
    #     async with resource_lock:
    #         async with aiosqlite.connect(self.__resource_db_path, timeout=30) as con:
    #             con.row_factory = aiosqlite.Row
    #             await con.execute('DELETE FROM resources WHERE pid == ?', (os.getpid(),))
    #             await con.commit()
    #     if self.__local_negotiator:
    #         await self.__local_negotiator.send_sync_event()

    async def run_task(self, task: InvocationJob, report_to: AddressChain):
        if self.__stopped:
            raise WorkerNotAvailable()
        self.__logger.debug(f'locks are {self.__task_changing_state_lock.locked()}, {self.__resource_db_lock.locked_by_me()}')
        async with self.__task_changing_state_lock, self.__resource_db_lock:
            self.__logger.debug('resource+task_change_state locks aquired')
            # we must ensure picking up and finishing tasks is in critical section
            # among all the local worker processes - hence this resource_db_lock which is file lock
            assert len(task.args()) > 0
            if self.__running_process is not None:
                raise AlreadyRunning('Task already in progress')

            # # check resources
            # self.__logger.debug(repr(self.__my_resources))
            # # await self.check_update_resources(already_locked=True)  # update our resources
            # task_resource_requirements = task.requirements().to_min_worker_resources()
            # self.__logger.debug(repr(self.__my_resources))
            # self.__logger.debug(repr(task_resource_requirements))
            # if self.__my_resources < task_resource_requirements:
            #     raise NotEnoughResources()

            # prepare logging
            self.__logger.info(f'running task {task}')

            # save external files
            self.__extra_files_base_dir = None
            extra_files_map: Dict[str, str] = {}
            if len(task.extra_files()) > 0:
                self.__extra_files_base_dir = tempfile.mkdtemp(prefix='lifeblood_efs_')  # TODO: add base temp dir to config
                self.__logger.debug(f'creating extra file temporary dir at {self.__extra_files_base_dir}')
            for exfilepath, exfiledata in task.extra_files().items():
                self.__logger.info(f'saving extra job file {exfilepath}')
                exfilepath_parts = exfilepath.split('/')
                tmpfilepath = os.path.join(self.__extra_files_base_dir, *exfilepath_parts)
                os.makedirs(os.path.dirname(tmpfilepath), exist_ok=True)
                with open(tmpfilepath, 'w' if isinstance(exfiledata, str) else 'wb') as f:
                    f.write(exfiledata)
                extra_files_map[exfilepath] = tmpfilepath

            # check args for extra file references
            if len(task.extra_files()) > 0:
                args = []
                for arg in task.args():
                    if isinstance(arg, str) and arg.startswith(':/') and arg[2:] in task.extra_files():
                        args.append(extra_files_map[arg[2:]])
                    else:
                        args.append(arg)
            else:
                args = task.args()

            try:
                if task.environment_resolver_arguments() is None:
                    config = get_config('worker')
                    env = environment_resolver.get_resolver(config.get_option_noasync('default_env_wrapper.name', 'TrivialEnvironmentResolver'))\
                        .get_environment(config.get_option_noasync('default_env_wrapper.arguments', {}))
                else:
                    env = task.environment_resolver_arguments().get_environment()
            except environment_resolver.ResolutionImpossibleError as e:
                self.__logger.error(f'cannot run the task: Unable to resolve environment: {str(e)}')
                raise

            # TODO: resolver args get_environment() acually does resolution so should be renamed to like resolve_environment()
            #  Environment's resolve() actually just expands and merges everything, so naming it "resolve" is misleading next to EnvironmentResolver

            env = task.env().resolve(env)

            env.prepend('PYTHONPATH', self.__rt_module_dir)
            env['LIFEBLOOD_RUNTIME_IID'] = task.invocation_id()
            env['LIFEBLOOD_RUNTIME_TID'] = task.task_id()
            env['LIFEBLOOD_RUNTIME_SCHEDULER_ADDR'] = self.__local_invocation_server_address_string
            for aname, aval in task.attributes().items():
                if aname.startswith('_'):  # skip attributes starting with _
                    continue
                # TODO: THINK OF A BETTER LOGIC !
                if isinstance(aval, (str, int, float)):
                    env[f'LBATTR_{aname}'] = str(aval)

            # env['LBATTRS_JSON'] = json.dumps(dict(task.attributes()))
            if self.__extra_files_base_dir is not None:
                env['LB_EF_ROOT'] = self.__extra_files_base_dir
            try:
                #with open(self.get_log_filepath('output', task.invocation_id()), 'a') as stdout:
                #    with open(self.get_log_filepath('error', task.invocation_id()), 'a') as stderr:
                # TODO: proper child process priority adjustment should be done, for now it's implemented in constructor.
                self.__running_process_start_time = time.time()

                if os.path.isabs(args[0]):
                    bin_path = args[0]
                else:
                    bin_path = shutil.which(args[0], path=env.get('PATH'))
                if bin_path is None:
                    raise ProcessInitializationError(f'"{args[0]}" was not found. Check environment resolver arguments and system setup')

                self.__running_process: asyncio.subprocess.Process = await create_process(args, env, os.path.dirname(bin_path))
            except Exception as e:
                self.__logger.exception('task creation failed with error: %s' % (repr(e),))
                raise

            self.__running_task = task
            self.__where_to_report = report_to
            self.__running_awaiter = asyncio.create_task(self._awaiter())
            self.__running_task_progress = 0
            if self.__worker_id is not None:  # TODO: gracefully handle connection fails here \/
                assert self.__pool_address is not None
                with WorkerPoolControlClient.get_worker_pool_control_client(self.__pool_address, self.__message_processor) as wpclient:  # type: WorkerPoolControlClient
                    await wpclient.report_state(self.__worker_id, WorkerState.BUSY)

            # # now adjust shared resource db:
            # # we are still in the resource lock
            # self.__my_resources -= task_resource_requirements
            # assert self.__my_resources.is_valid(), f'bad resources {repr(self.__my_resources)}'
            # await self._inform_scheduler_about_resources()
            # await self._set_local_resource_usage(task_resource_requirements, already_locked=True)

    # callback awaiter
    async def _awaiter(self):
        stdout_path = self.get_log_filepath('output', self.__running_task.invocation_id())
        stderr_path = self.get_log_filepath('error', self.__running_task.invocation_id())
        os.makedirs(os.path.dirname(stdout_path), exist_ok=True)
        os.makedirs(os.path.dirname(stderr_path), exist_ok=True)
        async with aiofiles.open(stdout_path, 'wb') as stdout:
            async with aiofiles.open(stderr_path, 'wb') as stderr:
                async def _flush():
                    await asyncio.sleep(1)  # ensure to flush every 1 second
                    await stdout.flush()
                    await stderr.flush()

                await stdout.write(datetime.datetime.now().strftime('[SYS][%d.%m.%y %H:%M:%S] task initialized\n').encode('UTF-8'))
                tasks_to_wait = {}
                try:
                    rout_task = asyncio.create_task(self.__running_process.stdout.readline())
                    rerr_task = asyncio.create_task(self.__running_process.stderr.readline())
                    done_task = asyncio.create_task(self.__running_process.wait())
                    flush_task = asyncio.create_task(_flush())
                    tasks_to_wait = {rout_task, rerr_task, done_task, flush_task}
                    while len(tasks_to_wait) != 0:
                        done, tasks_to_wait = await asyncio.wait(tasks_to_wait, return_when=asyncio.FIRST_COMPLETED)
                        if rout_task in done:
                            str = rout_task.result()
                            progress = self.__running_task.match_stdout_progress(str)
                            if progress is not None:
                                self.__running_task_progress = progress
                            if str != b'':  # this can only happen at eof
                                await stdout.write(datetime.datetime.now().strftime('[OUT][%H:%M:%S] ').encode('UTF-8') + str)
                                rout_task = asyncio.create_task(self.__running_process.stdout.readline())
                                tasks_to_wait.add(rout_task)
                        if rerr_task in done:
                            str = rerr_task.result()
                            progress = self.__running_task.match_stderr_progress(str)
                            if progress is not None:
                                self.__running_task_progress = progress
                            if str != b'':  # this can only happen at eof
                                await asyncio.gather(stderr.write(datetime.datetime.now().strftime('[ERR][%H:%M:%S] ').encode('UTF-8') + str),
                                                     stdout.write(datetime.datetime.now().strftime('[ERR][%H:%M:%S] ').encode('UTF-8') + str))
                                rerr_task = asyncio.create_task(self.__running_process.stderr.readline())
                                tasks_to_wait.add(rerr_task)
                        if flush_task in done and not done_task.done():
                            flush_task = asyncio.create_task(_flush())
                            tasks_to_wait.add(flush_task)
                    await stdout.write(datetime.datetime.now().strftime('[SYS][%d.%m.%y %H:%M:%S] task finished\n').encode('UTF-8'))
                except asyncio.CancelledError:
                    self.__logger.debug('task awaiter was cancelled')
                    for task in tasks_to_wait:
                        task.cancel()
                    raise
                finally:
                    # report to the pool
                    if self.__worker_id is not None:
                        try:
                            assert self.__pool_address is not None
                            with WorkerPoolControlClient.get_worker_pool_control_client(self.__pool_address, self.__message_processor) as wpclient:  # type: WorkerPoolControlClient
                                await wpclient.report_state(self.__worker_id, WorkerState.IDLE)
                        except (Exception, asyncio.CancelledError):
                            self.__logger.error('failed to report task cancellation to worker pool. stopping worker')
                            self.stop()

        await self.__running_process.wait()
        await self.task_finished()

    def is_task_running(self) -> bool:
        return self.__running_task is not None

    def running_invocation(self) -> Optional[InvocationJob]:
        return self.__running_task

    async def deliver_invocation_message(self, destination_invocation_id: int, destination_addressee: str, source_invocation_id: Optional[int], message_body: bytes, addressee_timeout: float = 90.0):
        """
        deliver message to task
        """
        while True:
            # while we wait - invocation MAY change.
            running_invocation = self.running_invocation()
            if running_invocation is None or destination_invocation_id != running_invocation.invocation_id():
                raise InvocationMessageWrongInvocationId()

            while destination_addressee not in self.__worker_task_comm_queues:
                wait_start_timestamp = time.time()
                await asyncio.sleep(0.05)  # we MOST LIKELY are already waiting for this, so timeout occurs
                addressee_timeout -= time.time() - wait_start_timestamp
                if addressee_timeout <= 0:
                    raise InvocationMessageAddresseeTimeout()
                # important to keep checking if invocation was changed,
                # and important to have no awaits (no interruptions) between check and enqueueing
                running_invocation = self.running_invocation()
                if running_invocation is None or destination_invocation_id != running_invocation.invocation_id():
                    raise InvocationMessageWrongInvocationId()

            queue = self.__worker_task_comm_queues[destination_addressee]

            if not queue.empty():
                # need to return control to loop in case 2 deliver_invocation_message calls happen to happen at the same time,
                # and one is stuck in the loop of upper while being satisfied, but queue not empty already
                await asyncio.sleep(0.01)
                continue
            queue.put_nowait((source_invocation_id, message_body))
            queue.put_nowait(())
            break

    async def worker_task_addressee_wait(self, addressee: str, timeout: float = 30) -> Tuple[int, bytes]:
        """
        wait for a data message to addressee to be delivered

        :returns: sender invocation id, message body
        """
        if addressee not in self.__worker_task_comm_queues:
            self.__worker_task_comm_queues[addressee] = asyncio.Queue()
        try:
            value = await asyncio.wait_for(self.__worker_task_comm_queues[addressee].get(), timeout=timeout)
            assert self.__worker_task_comm_queues[addressee].get_nowait() == ()
            # this way above we ensure one single deliver_task deliver to one single addressee_wait
        finally:
            self.__worker_task_comm_queues.pop(addressee)

        return value

    def is_stopping(self) -> bool:
        """
        returns True is stop was called on worker,
        so worker is closed or in the process of closing
        """
        return self.__stopped

    async def cancel_task(self):
        async with self.__task_changing_state_lock:
            if self.__running_process is None:
                return
            self.__logger.info('cancelling running task')
            self.__running_awaiter.cancel()
            cancelling_awaiter = self.__running_awaiter
            self.__running_awaiter = None

            await kill_process_tree(self.__running_process)

            self.__running_process._transport.close()  # sometimes not closed straight away transport ON EXIT may cause exceptions in __del__ that event loop is closed

            # report to scheduler that cancel was a success
            self.__logger.info(f'reporting cancel back to {self.__where_to_report}')
            try:
                with SchedulerWorkerControlClient.get_scheduler_control_client(self.__where_to_report, self.__message_processor) as client:  # type: SchedulerWorkerControlClient
                    await client.report_task_canceled(self.__running_task,
                                                      self.get_log_filepath('output', self.__running_task.invocation_id()),
                                                      self.get_log_filepath('error', self.__running_task.invocation_id()))
            except Exception as e:
                self.__logger.exception(f'could not report cuz of {e}')
            except:
                self.__logger.exception('could not report cuz i have no idea')
            # end reporting

            await self.delete_logs(self.__running_task.invocation_id())

            self.__running_task = None
            self.__worker_task_comm_queues = {}
            self.__running_process = None
            self.__where_to_report = None
            self.__running_task_progress = None
            await self._cleanup_extra_files()

            await asyncio.wait((cancelling_awaiter,))  # ensure everything is done before we proceed

            # # update resources
            # async with self.__resource_db_lock:
            #     await self._unset_local_resource_usage(already_locked=True)
            #     # self.__my_resources += task_used_resources  # not needed cuz check_update_resources will do it anyway
            #     # sanity_check_resources = copy.copy(self.__my_resources)
            #     await self.check_update_resources(already_locked=True)  # this also reports changed resources to scheruler
            #     # if self.__my_resources != sanity_check_resources:
            #     #     # no need to panic here - this can happen if resource change event was missed or was not yet processed

            # stop ourselves if we are a small task helper
            if self.__singleshot:
                self.stop()

    def task_status(self) -> Optional[float]:
        return self.__running_task_progress

    async def task_finished(self):
        """
        is called when current process finishes
        :return:
        """
        async with self.__task_changing_state_lock:
            if self.__running_process is None:
                self.__logger.warning('task_finished called, but there is no running task. This can only normally happen if a task_cancel happened the same moment as finish.')
                return
            self.__logger.info('task finished')
            process_exit_code = await self.__running_process.wait()
            self.__running_task.finish(process_exit_code, time.time() - self.__running_process_start_time)

            # report to scheduler
            self.__logger.info(f'reporting done back to {self.__where_to_report}')
            try:
                with SchedulerWorkerControlClient.get_scheduler_control_client(self.__where_to_report, self.__message_processor) as client:  # type: SchedulerWorkerControlClient
                    await client.report_task_done(self.__running_task,
                                                  self.get_log_filepath('output', self.__running_task.invocation_id()),
                                                  self.get_log_filepath('error', self.__running_task.invocation_id()))
            except Exception as e:
                self.__logger.exception(f'could not report cuz of {e}')
            except:
                self.__logger.exception('could not report cuz i have no idea')
            # end reporting

            await self.delete_logs(self.__running_task.invocation_id())

            self.__where_to_report = None
            self.__running_task = None
            self.__worker_task_comm_queues = {}
            self.__running_process = None
            self.__previous_notrunning_awaiter = self.__running_awaiter  # this is JUST so task is not GCd
            self.__running_awaiter = None  # TODO: lol, this function can be called from awaiter, and if we hand below - awaiter can be gcd, and it's all fucked
            self.__running_task_progress = None
            await self._cleanup_extra_files()

            # # update resources
            # async with self.__resource_db_lock:
            #     await self._unset_local_resource_usage(already_locked=True)
            #     # self.__my_resources += task_used_resources  # not needed cuz check_update_resources will do it anyway
            #     # sanity_check_resources = copy.copy(self.__my_resources)
            #     await self.check_update_resources(already_locked=True)  # this also reports changed resources to scheruler
            #     # if self.__my_resources != sanity_check_resources:
            #     #     # no need to panic here - this can happen if resource change event was missed or was not yet processed

            # stop ourselves if we are a small task helper
            if self.__singleshot:
                self.stop()

    async def _cleanup_extra_files(self):
        """
        cleanup extra files transfered with the task
        :return:
        """
        if self.__extra_files_base_dir is None:
            return
        try:
            shutil.rmtree(self.__extra_files_base_dir)
        except:
            self.__logger.exception('could not cleanup extra files')

    #
    # simply ping scheduler once in a while
    async def scheduler_pinger(self):
        """
        ping scheduler once in a while. if it misses too many pings - close worker and wait for new broadcasts
        :return:
        """

        async def _reintroduce_ourself():
            for attempt in range(5):
                self.__logger.debug(f'trying to reintroduce myself, attempt: {attempt + 1}')
                try:
                    with SchedulerWorkerControlClient.get_scheduler_control_client(self.__scheduler_addr, self.__message_processor) as client:  # type: SchedulerWorkerControlClient
                        assert self.__my_addr is not None
                        addr = self.__my_addr
                        self.__logger.debug('saying bye')
                        await client.say_bye(addr)
                        self.__logger.debug('cancelling task')
                        await self.cancel_task()
                        self.__logger.debug('saying hello')
                        self.__scheduler_db_uid = await client.say_hello(addr, self.__worker_type, self.__my_resources)
                        self.__logger.debug('reintroduce done')
                    break
                except Exception:
                    self.__logger.exception('failed to reintroduce myself. sleeping a bit and retrying')
                    await asyncio.sleep(10)
            else:  # failed to reintroduce. consider that something is wrong with the network, stop
                self.__logger.error('failed to reintroduce myself. assuming network problems, exiting')
                self.stop()

        exit_wait = asyncio.create_task(self.__components_stop_event.wait())
        while True:
            done, pend = await asyncio.wait((exit_wait, ), timeout=self.__ping_interval, return_when=asyncio.FIRST_COMPLETED)
            if exit_wait in done:
                await exit_wait
                break
            #await asyncio.sleep(self.__ping_interval)
            if self.__ping_missed_threshold == 0:
                continue
            # Here we are locking to prevent unexpected task state changes while checking for state inconsistencies
            async with self.__task_changing_state_lock:
                try:
                    self.__logger.debug('pinging scheduler')
                    with SchedulerWorkerControlClient.get_scheduler_control_client(self.__scheduler_addr, self.__message_processor) as client:  # type: SchedulerWorkerControlClient
                        result = await client.ping(self.__my_addr)
                    self.__logger.debug(f'scheduler pinged: sees me as {result}')
                except MessageTransferError as mte:
                    self.__logger.error('ping message delivery failed')
                    result = None
                except Exception as e:
                    self.__logger.exception('unexpected exception happened')
                    result = None
                task_running = self.is_task_running()

            if result is None:  # this means EOF
                self.__ping_missed += 1
                self.__logger.info(f'server ping missed. total misses: {self.__ping_missed}')
            if self.__ping_missed >= self.__ping_missed_threshold:
                # assume scheruler down, drop everything and look for another scheruler
                self.stop()
                return

            if result in (WorkerState.OFF, WorkerState.UNKNOWN):
                # something is wrong, lets try to reintroduce ourselves.
                # Note that we can be sure that there cannot be race conditions here:
                # pinger starts working always AFTER hello, OR it saz hello itself.
                # and scheduler will immediately switch worker state on hello, so ping coming after confirmed hello will ALWAYS get newer state
                self.__logger.warning(f'scheduler replied it thinks i\'m {result.name}. canceling tasks if any and reintroducing myself')
                await _reintroduce_ourself()
            elif result == WorkerState.BUSY and not task_running:
                # Note: the order is:
                # - sched sets worker to INVOKING
                # - shced sends "task"
                # - worker receives task, sets is_task_running
                # - worker answers to sched
                # - sched sets worker to BUSY
                # and when finished:
                # - worker reports done             |
                # - sched sets worker to IDLE       | under __task_changing_state_lock
                # - worker unsets is_task_running   |
                # so there is no way it can be not task_running AND sched state busy.
                # if it is - it must be an error
                self.__logger.warning(f'scheduler replied it thinks i\'m BUSY, but i\'m free, so something is inconsistent. resolving by reintroducing myself')
                await _reintroduce_ourself()
            elif result == WorkerState.IDLE and task_running:
                # Note from scheme above - this is not possible,
                #  the only period where scheduler can think IDLE while is_task_running set is in __task_changing_state_lock-ed area
                #  but we aquired sched state and our is_task_running above inside that __task_changing_state_lock
                self.__logger.warning(f'scheduler replied it thinks i\'m IDLE, but i\'m doing a task, so something is inconsistent. resolving by reintroducing myself')
                await _reintroduce_ourself()
            elif result == WorkerState.ERROR:
                # currently the only way it can be error is because of shitty network
                # ideally here we would check ourselves
                # but there's nothing to check right now
                self.__logger.warning('scheduler replied it thinks i\'m ERROR, but i\'m doing fine. probably something is wrong with the network. waiting for scheduler to resolve the problem')
                # no we don't reintroduce - error state on scheduler side just means he won't give us tasks for now
                # and since error is most probably due to network - it will either resolve itself, or there is no point reintroducing if connection cannot be established anyway
            elif result is not None:
                self.__ping_missed = 0


async def main_async(worker_type=WorkerType.STANDARD,
                     child_priority_adjustment: ProcessPriorityAdjustment = ProcessPriorityAdjustment.NO_CHANGE,
                     singleshot: bool = False, worker_id: Optional[int] = None, pool_address=None, noloop=False):
    """
    listen to scheduler broadcast in a loop.
    if received - create the worker and work
    if worker cannot ping the scheduler a number of times - it stops
    and listenting for broadcast starts again
    :return: Never!
    """
    graceful_closer_no_reentry = False

    def graceful_closer(*args):
        nonlocal graceful_closer_no_reentry
        if graceful_closer_no_reentry:
            print('DOUBLE SIGNAL CAUGHT: ALREADY EXITING')
            return
        graceful_closer_no_reentry = True
        logging.get_logger('worker').info('SIGINT/SIGTERM caught')
        nonlocal noloop
        noloop = True
        stop_event.set()
        if worker is not None:
            worker.stop()

    noasync_do_close = False

    def noasync_windows_graceful_closer_event(*args):
        nonlocal noasync_do_close
        noasync_do_close = True

    async def windows_graceful_closer():
        while not noasync_do_close:
            await asyncio.sleep(1)
        graceful_closer()

    worker = None
    stop_event = asyncio.Event()
    win_signal_waiting_task = None
    try:
        asyncio.get_event_loop().add_signal_handler(signal.SIGINT, graceful_closer)
        asyncio.get_event_loop().add_signal_handler(signal.SIGTERM, graceful_closer)
    except NotImplementedError:  # solution for windows
        signal.signal(signal.SIGINT, noasync_windows_graceful_closer_event)
        signal.signal(signal.SIGBREAK, noasync_windows_graceful_closer_event)
        win_signal_waiting_task = asyncio.create_task(windows_graceful_closer())

    config = get_config('worker')
    logger = logging.get_logger('worker')
    if await config.get_option('worker.listen_to_broadcast', True):
        stop_task = asyncio.create_task(stop_event.wait())
        while True:
            logger.info('listening for scheduler broadcasts...')
            broadcast_task = asyncio.create_task(await_broadcast('lifeblood_scheduler'))
            done, _ = await asyncio.wait((broadcast_task, stop_task), return_when=asyncio.FIRST_COMPLETED)
            if stop_task in done:
                broadcast_task.cancel()
                logger.info('broadcast listening cancelled')
                break
            assert broadcast_task.done()
            message = await broadcast_task
            scheduler_info = json.loads(message)
            logger.debug('received', scheduler_info)
            addr = AddressChain(scheduler_info['message_address'])
            try:
                worker = await create_worker(addr, child_priority_adjustment=child_priority_adjustment, worker_type=worker_type, singleshot=singleshot, worker_id=worker_id, pool_address=pool_address)
            except Exception:
                logger.exception('could not start the worker')
            else:
                await worker.wait_till_stops()
                logger.info('worker quited')
            if noloop:
                break
    else:
        logger.info('boradcast listening disabled')
        while True:
            addr = AddressChain(await config.get_option('worker.scheduler_address', get_default_addr()))
            logger.debug(f'using {addr}')
            try:
                worker = await create_worker(addr, child_priority_adjustment=child_priority_adjustment, worker_type=worker_type, singleshot=singleshot, worker_id=worker_id, pool_address=pool_address)
            except ConnectionRefusedError as e:
                logger.exception('Connection error', str(e))
                await asyncio.sleep(10)
                continue
            await worker.wait_till_stops()
            logger.info('worker quited')
            if noloop:
                break

    if win_signal_waiting_task is not None:  # this happens only on windows
        if not win_signal_waiting_task.done():
            win_signal_waiting_task.cancel()
    else:
        asyncio.get_event_loop().remove_signal_handler(signal.SIGINT)  # this seem to fix the bad signal fd error
        asyncio.get_event_loop().remove_signal_handler(signal.SIGTERM)  # my guess what happens is that loop closes, but signal handlers remain if not unsed


default_config = '''
[worker]
listen_to_broadcast = true

[default_env_wrapper]
## here you can uncomment lines below to specify your own default environment wrapper and default arguments
## this will only be used by invocation jobs that have NO environment wrappers specified
# name = TrivialEnvironmentResolver
# arguments = [ "project_name", "or", "config_name", "idunno", "maybe rez packages requirements?", [1,4,11] ]

[resources]
## here you can override resources that this machine has
## if you don't specify anything - resources will be detected automatically
## NOTE: automatic detection DOES NOT WORK FOR GPU yet, you have to specify it manually
# cpu_count = 32    # by default treated as the number of cores 
# cpu_mem = "128G"  # you can either specify int amount of bytes, or use string ending with one of "K" "M" "G" "T" "P" meaning Kilo, Mega, Giga, ... 
# gpu_count = 1     # by default treated as the number devices
# gpu_mem = "8G"    # you can either specify int amount of bytes, or use string ending with one of "K" "M" "G" "T" "P" meaning Kilo, Mega, Giga, ...
'''


def main(argv):
    # import signal
    # prev = None
    # def signal_handler(sig, frame):
    #     print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! You pressed Ctrl+C !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
    #     prev(sig, frame)
    #
    # prev = signal.signal(signal.SIGINT, signal_handler)
    import argparse
    parser = argparse.ArgumentParser('lifeblood worker', description='executes invocations from scheduler')
    parser.add_argument('--scheduler-address', help='manually specify scheduler to connect to. if not specified - by default worker will start listening to broadcasts from schedulers')
    parser.add_argument('--no-listen-broadcast', action='store_true', help='do not listen to scheduler\'s broadcast, use config')
    parser.add_argument('--no-loop', action='store_true', help='by default worker will return into the loop of waiting for scheduler every time it quits because of connection loss, or other errors. '
                                                               'but this flag will force worker to just completely quit instead')
    parser.add_argument('--singleshot', action='store_true', help='worker will pick one job and exit after that job is completed or cancelled. '
                                                                  'this is on by default when type=SCHEDULER_HELPER')
    parser.add_argument('--type', choices=('STANDARD', 'SCHEDULER_HELPER'), default='STANDARD')
    parser.add_argument('--id', help='integer identifier which worker should use when talking to worker pool')
    parser.add_argument('--pool-address', help='if this worker is a part of a pool - pool address. currently pool can only be on the same host')
    parser.add_argument('--priority', choices=tuple(x.name for x in ProcessPriorityAdjustment), default=ProcessPriorityAdjustment.NO_CHANGE.name, help='adjust child process priority')
    parser.add_argument('--generate-config-only', action='store_true', help='just generate initial config and exit. Note that existing config will NOT be overriden')

    args = parser.parse_args(argv)

    # check and create default config if none
    create_default_user_config_file('worker', default_config)

    if args.generate_config_only:
        return

    if args.type == 'STANDARD':
        wtype = WorkerType.STANDARD
    elif args.type == 'SCHEDULER_HELPER':
        wtype = WorkerType.SCHEDULER_HELPER
    else:
        raise NotImplementedError(f'worker type {args.type} is not yet implemented')

    priority_adjustment = [x for x in ProcessPriorityAdjustment if x.name == args.priority][0]  # there MUST be exactly 1 match

    global_logger = logging.get_logger('worker')

    # check and create default config if none
    create_default_user_config_file('worker', default_config)

    # check legality of the address
    paddr = AddressChain(args.pool_address)

    config = get_config('worker')
    if args.no_listen_broadcast:
        config.set_override('worker.listen_to_broadcast', False)
    if args.scheduler_address is not None:
        config.set_override('worker.listen_to_broadcast', False)
        saddr = AddressChain(args.scheduler_address)
        config.set_override('worker.scheduler_address', str(saddr))
    try:
        asyncio.run(main_async(wtype, child_priority_adjustment=priority_adjustment, singleshot=args.singleshot, worker_id=int(args.id) if args.id is not None else None, pool_address=paddr, noloop=args.no_loop))
    except KeyboardInterrupt:
        # if u see errors in pycharm around this area when running from scheduler -
        # it's because pycharm and most shells send SIGINTs to this child process on top of SIGINT that pool sends
        # this stuff above tries to suppress that double SIGINTing, but it's not 100% solution
        global_logger.warning('SIGINT caught where it wasn\'t supposed to be caught')
        global_logger.info('SIGINT caught. Worker is stopped now.')


if __name__ == '__main__':
    import sys
    main(sys.argv)



import asyncio
from ..component_base import ComponentBase
from ..enums import SchedulerMode

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # TODO: maybe separate a subset of scheduler's methods to smth like SchedulerData class, or idunno, for now no obvious way to separate, so having a reference back
    from .scheduler import Scheduler


class SchedulerComponentBase(ComponentBase):
    def __init__(self, scheduler: "Scheduler"):
        super().__init__()
        self.__stop_event = asyncio.Event()
        self.__main_task = None
        self.__wakeup_event = asyncio.Event()
        self.__scheduler = scheduler
        self.__mode = SchedulerMode.STANDARD

    @property
    def scheduler(self):
        return self.__scheduler

    @property
    def mode(self):
        return self.__mode

    @property
    def _poke_event(self):
        return self.__wakeup_event

    def poke(self):
        """
        poke pinger to interrupt sleep and continue pinging immediately
        """
        self.__wakeup_event.set()

    def _reset_poke_event(self):
        self.__wakeup_event.clear()

    def _main_task(self):
        raise NotImplementedError('override this with the main task')

    def _my_sleep(self):
        raise NotImplementedError('override this')

    def _my_wake(self):
        raise NotImplementedError('override this')

    def sleep(self):
        if self.__mode == SchedulerMode.DORMANT:
            return
        self.__mode = SchedulerMode.DORMANT
        self._my_sleep()
        self.scheduler._component_changed_mode(self, self.__mode)

    def wake(self):
        if self.__mode == SchedulerMode.STANDARD:
            return
        self.__mode = SchedulerMode.STANDARD
        self._my_wake()
        self.scheduler._component_changed_mode(self, self.__mode)



from contextlib import contextmanager
from .enums import WorkerState
from .net_messages.impl.tcp_simple_command_message_processor import TcpCommandMessageProcessor
from .net_messages.impl.clients import CommandJsonMessageClient
from .net_messages.address import AddressChain
from .net_messages.messages import Message
from .net_messages.impl.message_haldlers import CommandMessageHandlerBase


from typing import Optional, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from .simple_worker_pool import WorkerPool


class WorkerPoolMessageHandler(CommandMessageHandlerBase):
    def __init__(self, worker_pool: "WorkerPool"):
        super().__init__()
        self.__worker_pool = worker_pool

    def command_mapping(self):
        return {
            'worker.state_report': self._command_state_report
        }

    #
    # commands
    #

    async def _command_state_report(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
            worker_id: worker id as assigned by pool
            state: WorkerState
        returns keys:
            ok: ok is ok
        """
        state = WorkerState(args['state'])
        await self.__worker_pool._worker_state_change(args['worker_id'], state)
        await client.send_message_as_json({
            'ok': True
        })


class WorkerPoolMessageProcessor(TcpCommandMessageProcessor):
    def __init__(self, worker_pool: "WorkerPool", listening_address: Tuple[str, int], *, backlog=4096, connection_pool_cache_time=300):
        super().__init__(listening_address,
                         backlog=backlog,
                         connection_pool_cache_time=connection_pool_cache_time,
                         message_handlers=(WorkerPoolMessageHandler(worker_pool),))


#
# Client
#


class WorkerPoolControlClient:
    def __init__(self, client: CommandJsonMessageClient):
        self.__client = client

    @classmethod
    @contextmanager
    def get_worker_pool_control_client(cls, scheduler_address: AddressChain, processor: TcpCommandMessageProcessor) -> "WorkerPoolControlClient":
        with processor.message_client(scheduler_address) as message_client:
            yield WorkerPoolControlClient(message_client)

    async def report_state(self, worker_id: int, state: WorkerState):
        await self.__client.send_command('worker.state_report', {
            'worker_id': worker_id,
            'state': state.value
        })
        reply = await self.__client.receive_message()
        assert (await reply.message_body_as_json()).get('ok', False), 'something is not ok'



import re
from toml.encoder import _dump_str, TomlEncoder, InlineTableDict


class TomlFlatConfigEncoder(TomlEncoder):
    def __init__(self, _dict=dict):
        super(TomlFlatConfigEncoder, self).__init__(_dict)

    def dump_sections(self, o, sup):
        # print('in:', repr(o), repr(sup))

        retstr = ""
        if sup != "" and sup[-1] != ".":
            sup += '.'
        retdict = self._dict()
        arraystr = ""
        interesting_section = any(isinstance(x, self._dict) for x in o.values()) and any(not isinstance(x, self._dict) for x in o.values())
        for section in o:
            qsection = section
            if not re.match(r'^[A-Za-z0-9_-]+$', section):
                qsection = _dump_str(section)
            if not isinstance(o[section], dict):
                arrayoftables = False
                if isinstance(o[section], list):
                    for a in o[section]:
                        if isinstance(a, dict):
                            arrayoftables = True
                if arrayoftables:
                    for a in o[section]:
                        arraytabstr = "\n"
                        arraystr += "[[" + sup + qsection + "]]\n"
                        s, d = self.dump_sections(a, sup + qsection)
                        if s:
                            if s[0] == "[":
                                arraytabstr += s
                            else:
                                arraystr += s
                        while d:
                            newd = self._dict()
                            for dsec in d:
                                s1, d1 = self.dump_sections(d[dsec], sup +
                                                            qsection + "." +
                                                            dsec)
                                if s1:
                                    arraytabstr += ("[" + sup + qsection +
                                                    "." + dsec + "]\n")
                                    arraytabstr += s1
                                for s1 in d1:
                                    newd[dsec + "." + s1] = d1[s1]
                            d = newd
                        arraystr += arraytabstr
                else:
                    if o[section] is not None:
                        retstr += (qsection + " = " +
                                   str(self.dump_value(o[section])) + '\n')
            elif self.preserve and isinstance(o[section], InlineTableDict):
                retstr += (qsection + " = " +
                           self.dump_inline_table(o[section]))
            else:
                if interesting_section:
                    prefix = qsection
                    subrem = o[section]

                    def _helper(prefix, subrem):
                        nonlocal retstr
                        while subrem != {}:
                            subret, subrem = self.dump_sections(subrem, sup + prefix)
                            if subret:
                                retstr += '\n'.join(prefix + '.' + x for x in subret.strip().split('\n')) + '\n'
                            else:
                                # it seeeeems that in case subret == '' -
                                # then we can expect all keys of subrem to be already quoted....
                                # that's what dump_sections default implementation does
                                # so we don't need to quote subsection
                                for subsection in subrem:
                                    # qsubsection = subsection
                                    # if not re.match(r'^[A-Za-z0-9_-]+$', subsection):
                                    #     qsubsection = _dump_str(subsection)
                                    _helper(prefix + "." + subsection, subrem[subsection])
                                break
                    _helper(prefix, subrem)

                else:
                    retdict[qsection] = o[section]
        retstr += arraystr
        # print('out:', repr(retstr), repr(retdict))
        return retstr, retdict


if __name__ == '__main__':
    # not a test, but a demonstration
    import toml
    d = {'packages': {
        'houdini.py2': {
            '18.0.597': {
                'env': {
                    'PATH': {
                        'prepend': '/opt/hfs18.0.597/bin',
                        'append': 'ass'
                    }
                },
                'label': 'ass'
            },
            '18.5.408': {
                'env': {
                    'PATH': {
                        'prepend': '/opt/hfs18.5.408/bin'
                    },
                    'vata': [1, 2, 5]
                }
            }
        },
        'houdini.py3': {
            '18.5.499': {
                'env': {
                    'PATH': {
                        'prepend': '/opt/hfs18.5.499.py3/bin',
                        'append': 'me'
                    }
                }
            }
        },
            'bonker': {
                '18.5.499': {
                    'lobe.cabe': 2.3,
                    'iii': -1,
                    'env.shmenv': {
                        'PATH.SHMATH': {
                            'prepend': '/opt/hfs18.5.499.py3/bin',
                            'append': 'me'
                        },
                        'BOB.SHMOB': {
                            'prepend': 'you',
                            'append': 'me'
                        }
                    }
                }
            }
    }
    }
    r = toml.dumps(d, encoder=TomlFlatConfigEncoder())
    r2 = toml.dumps(d)
    print(r)
    print('\n\n============================\n\n')
    print(r2)



"""
environment wrapper is an object that produces runtime environment for an invocation
taking into account invocation's requirements and worker's machine specifics

For example you might want to at least pass the version of software you want to use together with
your invocation job

Or a more complicated environment wrapper would take in a whole set of required packages from invocation job
and produce an environment fitting those requirements

As I see the it, a freelancer or a studio would implement one specific to them environment wrapper
for all workers, not several different wrappers
"""
import asyncio
import os
import sys
import json
import inspect
import pathlib
import re
from copy import deepcopy
from semantic_version import Version, SimpleSpec
from types import MappingProxyType
from . import invocationjob, paths, logging
from .config import get_config
from .toml_coders import TomlFlatConfigEncoder
from .process_utils import oh_no_its_windows

from typing import Dict, Mapping, Optional, Type, Iterable


_resolvers: Dict[str, Type["BaseEnvironmentResolver"]] = {}  # this should be loaded from plugins


def _populate_resolvers():
    for k, v in dict(globals()).items():
        if not inspect.isclass(v) \
                or not issubclass(v, BaseEnvironmentResolver) \
                or v == BaseEnvironmentResolver \
                or v.__module__ != __name__:
            continue
        _resolvers[k] = v
    logging.get_logger('environment_resolver_registry').info('resolvers found:\n' + '\n'.join(f'\t{k}' for k in _resolvers))


def get_resolver(name: str) -> "BaseEnvironmentResolver":
    return _resolvers[name]()


class ResolutionImpossibleError(RuntimeError):
    pass


class EnvironmentResolverArguments:
    """
    this class objects specity requirements a task/invocation have for int's worker environment wrapper.
    """
    def __init__(self, resolver_name=None, arguments: Optional[Mapping] = None):
        """

        :param resolver_name: if None - treat as no arguments at all
        :param arguments:
        """
        if arguments is None:
            arguments = {}
        if resolver_name is None and len(arguments) > 0:
            raise ValueError('if name is None - no arguments are allowed')
        self.__resolver_name = resolver_name
        self.__args = arguments

    def name(self):
        return self.__resolver_name

    def set_name(self, name: str):
        self.__resolver_name = name

    def arguments(self):
        return MappingProxyType(self.__args)

    def add_argument(self, name: str, value):
        self.__args[name] = value

    def remove_argument(self, name: str):
        del self.__args[name]

    def get_resolver(self):
        return get_resolver(self.__resolver_name)

    def get_environment(self) -> "invocationjob.Environment":
        return get_resolver(self.name()).get_environment(self.arguments())

    def serialize(self) -> bytes:
        return json.dumps(self.__dict__).encode('utf-8')

    async def serialize_async(self):
        return await asyncio.get_running_loop().run_in_executor(None, self.serialize)

    @classmethod
    def deserialize(cls, data: bytes):
        wrp = EnvironmentResolverArguments(None)
        wrp.__dict__.update(json.loads(data.decode('utf-8')))
        return wrp

    @classmethod
    async def deserialize_async(cls, data: bytes):
        return await asyncio.get_running_loop().run_in_executor(None, cls.deserialize, data)


class BaseEnvironmentResolver:
    def get_environment(self, arguments: Mapping) -> "invocationjob.Environment":
        """
        this is the main reason for environment wrapper's existance.
        give it your specific arguments

        :param additional_env:
        :param arguments:
        :return:
        """
        raise NotImplementedError()


class TrivialEnvironmentResolver(BaseEnvironmentResolver):
    """
    trivial environment wrapper does nothing
    """
    def get_environment(self, arguments: dict) -> "invocationjob.Environment":
        env = invocationjob.Environment(os.environ)
        return env


class StandardEnvironmentResolver(BaseEnvironmentResolver):
    """
    will initialize environment based on requested software versions and it's own config
    will raise ResolutionImpossibleError if he doesn't know how to resolve given configuration

    example configuration:
    [packages.houdini."18.5.666"]
    env.PATH.prepend=[
        "/path/to/hfs/bin",
        "/some/other/path/dunno"
    ]
    env.PATH.append=[
        "/whatever/you/want/to/append"
    ]
    env.PYTHONPATH.prepend="/dunno/smth"
    """
    logger = logging.get_logger('environment resolver')

    def __init__(self):
        # attempt to locate config and autocreate if not found
        config = get_config('standard_environment_resolver')
        if len(config.loaded_files()) + len(config.broken_files()) == 0:
            config.set_toml_encoder_generator(TomlFlatConfigEncoder)
            self.logger.info('standard environment resolver is used, but no configuration found. auto generating configuration...')
            packages = self.autodetect_software()
            for pkgname, v in packages.items():
                for verstr in v.keys():
                    self.logger.info(f'found {pkgname} : {verstr}')
            config.set_option_noasync('packages', packages)
            self.logger.info(f'autogenerated config saved to {config.writeable_file()}')
        elif len(config.loaded_files()) == 0:  # so all sources are broken
            self.logger.error('environment resolver configs found, but all have errors! Aborting!')
            raise RuntimeError('all resolver configs are broken')

    def get_environment(self, arguments: Mapping) -> "invocationjob.Environment":
        """

        :param arguments: are expected to be in format of package_name: version_specification
                          like houdini
        :return:
        """
        config = get_config('standard_environment_resolver')
        packages = config.get_option_noasync('packages')
        if packages is None:
            raise ResolutionImpossibleError('no packages are configured')

        available_software = {k: {Version(v): rest for v, rest in packages[k].items()} for k in packages.keys()}

        resolved_versions = {}
        for package, spec_str in arguments.items():
            if not package.startswith('package.'):  # all package reqs start with package.
                continue
            package = package[len('package.'):]
            if package not in available_software:
                raise ResolutionImpossibleError(f'no configurations for package {package} found')
            resolved_versions[package] = SimpleSpec(spec_str).select(available_software[package].keys())
            if resolved_versions[package] is None:
                raise ResolutionImpossibleError(f'could not satisfy version requirements {spec_str} for package {package}')

        env = invocationjob.Environment(os.environ)
        for package, version in sorted(resolved_versions.items(), key=lambda x: available_software[x[0]][x[1]].get('priority', 50)):
            actions = available_software[package][version]
            for env_name, env_action in actions.get('env', {}).items():
                if not isinstance(env_action, Mapping):
                    env_action = {'set': env_action}
                if 'prepend' in env_action:
                    value = env_action['prepend']
                    if isinstance(value, str):
                        value = [value]
                    for part in reversed(value):
                        env.prepend(env_name, part)
                if 'append' in env_action:
                    value = env_action['append']
                    if isinstance(value, str):
                        value = [value]
                    for part in value:
                        env.append(env_name, part)
                if 'set' in env_action:
                    value = env_action['set']
                    if isinstance(value, list):
                        value = os.pathsep.join(value)
                    env[env_name] = value
        if 'user' in arguments:
            for uservar in ('USER', 'LOGNAME', 'USERNAME'):
                env[uservar] = arguments['user']
        if oh_no_its_windows and 'PYTHONIOENCODING' not in env:
            env['PYTHONIOENCODING'] = 'UTF-8'
        return env

    @classmethod
    def autodetect_software(cls, base_path: Optional[str] = None, software_to_detect: Optional[Iterable[str]] = None) -> dict:
        """
        scans common install locations, tries to detect some software
        (currently only works on houdini...)

        :return:
        """
        full_software_set = {'houdini', 'blender'}
        base = pathlib.Path(base_path) if base_path is not None else None
        software_to_detect = set(x.lower() for x in software_to_detect) if software_to_detect is not None else full_software_set
        all = {}
        if 'houdini' in software_to_detect:
            try:
                all.update(cls.autodetect_houdini(base))
            except Exception as e:
                cls.logger.exception('failed to detect houdini, skipping')
        if 'blender' in software_to_detect:
            try:
                all.update(cls.autodetect_blender(base))
            except Exception as e:
                cls.logger.exception('failed to detect blender, skipping')
        return all

    @classmethod
    def autodetect_houdini(cls, base_path: Optional[pathlib.Path] = None) -> dict:
        if sys.platform.startswith('linux'):
            base = base_path or pathlib.Path(r'/opt')
            hfs_prefix = pathlib.Path('')
        elif sys.platform.startswith('win'):
            base = base_path or pathlib.Path(r'C:\Program Files\Side Effects Software')
            hfs_prefix = pathlib.Path('')
        elif sys.platform.startswith('darwin'):
            base = base_path or pathlib.Path(r'/Applications/Houdini')
            hfs_prefix = pathlib.Path('Frameworks/Houdini.framework/Versions/Current/Resources')
        else:
            raise RuntimeError(f'unknown platform {sys.platform}')

        packages = {}

        houre = re.compile(r'^(?:[Hh]oudini|hfs)\s*(\d+\.\d+\.\d+)(?:\.py(\d+))?$')
        pyre = re.compile(r'^python(\d+)\.(\d+).*$')
        if base.exists():
            try:
                for houdir in base.iterdir():
                    if not houdir.exists() or not houdir.is_dir():
                        continue
                    match = houre.match(str(houdir.name))
                    if not match:
                        continue
                    hpy = match.group(2)
                    if hpy is None:  # so we don't see explicit python version
                        for file in (houdir/hfs_prefix/'python'/'bin').iterdir():
                            pymatch = pyre.match(str(file.name))
                            if not pymatch:
                                continue
                            hpy = pymatch.group(1)
                            break
                    if f'houdini.py{hpy}' not in packages:
                        packages[f'houdini.py{hpy}'] = {}
                    packages[f'houdini.py{hpy}'][match.group(1)] = {
                        'label': f'SideFX Houdini, with python version {hpy}',
                        'env': {
                            'PATH': {
                                'prepend': str(houdir/hfs_prefix/'bin')
                            }
                        }
                    }
            except PermissionError:
                cls.logger.error(f'no permissions to scan {base}, skipping')
        return packages

    @classmethod
    def autodetect_blender(cls, base_path: Optional[pathlib.Path] = None) -> dict:
        bases = []
        packages = {}
        if sys.platform.startswith('linux'):
            return {}  # TODO: implement? any standard way of installinb blender is with package manager, cannot cover that case
        elif sys.platform.startswith('win'):
            if base_path:
                bases.append(base_path)
            else:
                bases.append(pathlib.Path(r'C:\Program Files\Blender Foundation'))
                bases.append(base_path or pathlib.Path(r'C:\Program Files (x86)\Blender Foundation'))
            blere = re.compile(r'^(?:[Bb]lender)\s*(\d+\.\d+(?:.\d+)*)$')
            for base in bases:
                if not base.exists():
                    continue
                try:
                    for blendir in base.iterdir():
                        if not blendir.exists() or not blendir.is_dir():
                            continue
                        match = blere.match(str(blendir.name))
                        if not match:
                            continue
                        if 'blender' not in packages:
                            packages['blender'] = {}
                        packages['blender'][match.group(1)] = {
                            'label': 'Blender',
                            'env': {
                                'PATH': {
                                    'prepend': str(blendir)
                                }
                            }
                        }
                except PermissionError:
                    cls.logger.error(f'no permissions to scan {base}, skipping')
        elif sys.platform.startswith('darwin'):
            import subprocess
            # TODO: single stupid case covered only, not in a super nice way either...
            bases.append(base_path or pathlib.Path(r'/Applications/Blender.app/Contents/MacOS'))
            for base in bases:
                blenbin = base / 'blender'
                if blenbin.exists():
                    try:
                        out, _ = subprocess.Popen([blenbin, '--version'], stdout=subprocess.PIPE).communicate()
                    except:
                        cls.logger.warning(f' could not launch {(base / "blender")} to check version')
                        continue
                    out = out.decode('utf-8')
                    match = re.match(r'[Bb]lender\s*(\d+\.\d+)', out.splitlines(False)[0])
                    if not match:
                        continue
                    packages['blender'][match.group(1)] = {
                        'label': 'Blender',
                        'env': {
                            'PATH': {
                                'prepend': str(base)
                            }
                        }
                    }
        else:
            raise RuntimeError(f'unknown platform {sys.platform}')

        return packages


_populate_resolvers()


def main(args):
    """
    autodetect things for standard environment resolver
    """
    import argparse
    parser = argparse.ArgumentParser('environment_resolver', description='generates standard environment resolver config\n'
                                                                         'currently only Houdini is supported')
    subparser = parser.add_subparsers(title='command', required=True, dest='command')

    cmdparser = subparser.add_parser('generate', description='generate configuration file in user dir')
    cmdparser.add_argument('--basepath', '-p', help='optional comma (,) separated base path list to search software in')
    cmdparser.add_argument('--output', '-o', help='save config to this file instead of default config location')
    cmdparser.add_argument('--override', help='if set old config will be completely overwritten by a new one. '
                                              'Otherwise only new entries will be added to the config, keeping existing'
                                              'entries intact', action='store_true')

    schparser = subparser.add_parser('scan', description='scan for software but do NOT generate config files')
    schparser.add_argument('--basepath', '-p', help='optional comma (,) separated base path list to search software in')

    opts = parser.parse_args(args)

    packages = StandardEnvironmentResolver.autodetect_software()
    if opts.basepath:
        for basepath in opts.basepath.split(','):
            packages.update(StandardEnvironmentResolver.autodetect_software(basepath))

    if opts.command == 'generate':
        config = get_config('standard_environment_resolver')
        if opts.output:
            config.override_config_save_location(opts.output)
        config.set_toml_encoder_generator(TomlFlatConfigEncoder)
        logger = logging.get_logger('environment resolver')
        logger.info('standard environment resolver is used, but no configuration found. auto generating configuration...')
        for pkgname, v in packages.items():
            for verstr in v.keys():
                logger.info(f'found {pkgname} : {verstr}')

        if opts.override:  # do full config override
            config.set_option_noasync('packages', packages)
            logger.info(f'autogenerated config saved to {config.writeable_file()}')
        else:  # if not full override - only add new entries, no touch existing ones
            conf_packages = deepcopy(config.get_option_noasync('packages', {}))
            needs_resaving = False
            for package_name, package_data in packages.items():
                if package_name not in conf_packages:
                    conf_packages[package_name] = package_data
                    needs_resaving = True
                    continue
                for ver, meta in package_data.items():
                    if ver in conf_packages[package_name]:
                        continue
                    conf_packages[package_name][ver] = meta
                    needs_resaving = True
            if needs_resaving:
                config.set_option_noasync('packages', conf_packages)

    elif opts.command == 'scan':
        for pkgname, stuff in packages.items():
            print(f'{pkgname}:')
            for ver, meta in stuff.items():
                print(f'\t{ver}' + (f' ({meta.get("label", "")})' if 'label' in meta else ''))
                print('\n'.join(f'\t\t{k}: {v}' for k, v in meta.items() if k != 'label'))


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]) or 0)



# TODO: this should be deployed automatically to this particular houdini folders
#  and to other places form from some central location AUTOMATICALLY

"""
this module is supposed to be accessible for tasks ran by worker.
this module should be kept minimal, with only standard modules and maximum compatibility
"""
from __future__ import print_function, absolute_import
import pickle
import socket
import struct
from getpass import getuser
from .nethelpers import recv_string, recv_exactly, send_string
from .query import scheduler_client, Task


class TaskSpawn(object):
    """
    this class is a pickle compatible shrunk copy of lifeblood.taskspawn.TaskSpawn
    keep it up-to-date
    and keep it 2-3 compatible!!
    """
    def __init__(self, name, source_invocation_id, env_args, task_attributes):
        self.__name = name
        self.__attributes = dict(task_attributes or {})
        self.__env_args = env_args
        self.__forced_node_task_id_pair = None
        self.__from_invocation_id = source_invocation_id
        self.__output = 'spawned'
        self._create_as_spawned = True
        self.__extra_groups = []
        self.__default_priority = None

    def create_as_spawned(self):
        return self._create_as_spawned

    def force_set_node_task_id(self, node_id, task_id):
        self.__forced_node_task_id_pair = (node_id, task_id)

    def forced_node_task_id(self):
        return self.__forced_node_task_id_pair

    def source_invocation_id(self):
        return self.__from_invocation_id

    def set_node_output_name(self, new_name):
        self.__output = new_name

    def node_output_name(self):
        return self.__output

    def name(self):
        return self.__name

    def default_priority(self):
        """
        This priority will be used only in case this task requires a default group creation
        If this task has nonempty list of groups to be assigned to - this default priority is

        :return: default priority
        """
        return self.__default_priority

    def add_extra_group_name(self, group_name):
        self.__extra_groups.append(group_name)

    def extra_group_names(self):
        return self.__extra_groups

    def set_default_priority(self, priority):
        """
        This priority will be used only in case this task requires a default group creation
        If this task has nonempty list of groups to be assigned to - this default priority is
        """
        self.__default_priority = priority

    def set_name(self, name):
        self.__name = name

    def set_attribute(self, attr_name, attr_value):
        self.__attributes[attr_name] = attr_value

    def remove_attribute(self, attr_name):
        del self.__attributes[attr_name]

    def attribute_value(self, attr_name):
        return self.__attributes.get(attr_name, None)

    def _attributes(self):
        return self.__attributes

    def environment_arguments(self):
        return self.__env_args

    def set_environment_resolver(self, resolver_name, resolver_arguments):
        self.__env_args = EnvironmentResolverArguments(resolver_name, resolver_arguments)

    def serialize(self):
        return pickle.dumps(self)


class EnvironmentResolverArguments:
    """
    this is a copy of envirionment_resolver.EnvironmentResolverArguments class purely for pickling
    """
    def __init__(self, resolver_name=None, arguments=None):
        if arguments is None:
            arguments = {}
        if resolver_name is None and len(arguments) > 0:
            raise ValueError('if name is None - no arguments are allowed')
        self.__resolver_name = resolver_name
        self.__args = arguments


class NewTask(TaskSpawn):
    def __init__(self, name, node_id, scheduler_addr, env_args, task_attributes, priority=50.0):
        super(NewTask, self).__init__(name, None, env_args, task_attributes)
        self.__scheduler_addr = scheduler_addr
        self.set_node_output_name('main')
        self.force_set_node_task_id(node_id, None)
        self._create_as_spawned = False
        self.set_default_priority(priority)

    def submit(self):
        addr, sport = self.__scheduler_addr
        port = int(sport)
        sock = socket.create_connection((addr, port), timeout=30)
        sock.sendall(b'\0\0\0\0')
        data = self.serialize()

        send_string(sock, 'spawn')
        sock.sendall(struct.pack('>Q', len(data)))
        sock.sendall(data)
        status, is_not_null, task_id = struct.unpack('>I?Q', sock.recv(13))
        if status != 0:
            raise RuntimeError('scheduler failed to create task')
        assert is_not_null
        return Task(self.__scheduler_addr, task_id)


def create_task(name, node_id_or_name,
                scheduler_addr,
                attributes=None,
                env_resolver_name='StandardEnvironmentResolver', env_arguments=None,
                priority=50.0):

    if attributes is None:
        attributes = {}
    if env_arguments is None:
        env_arguments = {'user': getuser()}
    env_res = EnvironmentResolverArguments(env_resolver_name, env_arguments)
    if isinstance(node_id_or_name, str):
        addr, sport = scheduler_addr
        port = int(sport)
        with scheduler_client(addr, port) as client:
            nodeids = client.find_nodes_by_name(node_id_or_name)

        if len(nodeids) == 0:
            raise RuntimeError('lifeblood graph does not have node named %s' % node_id_or_name)
        node_id = nodeids[0]
    else:
        assert isinstance(node_id_or_name, int)
        node_id = node_id_or_name

    task = NewTask(name, node_id, scheduler_addr, env_args=env_res, task_attributes=attributes, priority=priority)
    return task



from copy import copy
from lifeblood.basenode import BaseNode
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
import math

from typing import Iterable


def node_class():
    return FramerangeSplitter


class FramerangeSplitter(BaseNode):

    @classmethod
    def label(cls) -> str:
        return 'attribute splitter'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'attribute', 'splitter', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'attribute_splitter'

    @classmethod
    def description(cls) -> str:
        return 'splits task\'s attribute value into chunks of "chunk size" elements\n' \
               'Works ONLY on array/list attributes\n' \
               'NOTE that task wil ALWAYS be splitted, even if just in 1 piece\n' \
               '\n' \
               'for ex. attribute "something" of value [3,4,5,6,7,8,9,10,11,12,13]\n' \
               'with chunk size of 4\n' \
               'will split task into 3 tasks with attributes "something" equals to:\n' \
               '[3,4,5,6]  [7,8,9,10]  [11,12,13]  respectivelly'

    def __init__(self, name):
        super(FramerangeSplitter, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            mode_param = ui.add_parameter('split type', 'mode', NodeParameterType.STRING, 'list') \
                            .add_menu((('Split List', 'list'), ('Split Range', 'range')))
            ui.add_separator()
            ui.add_parameter('attribute name', 'attribute value to split', NodeParameterType.STRING, 'frames').append_visibility_condition(mode_param, '==', 'list')
            ui.add_parameter('chunk size', 'chunk size', NodeParameterType.INT, 10).append_visibility_condition(mode_param, '==', 'list')

            range_mode_param = ui.add_parameter('range mode', 'Range Mode', NodeParameterType.INT, 0) \
                                    .add_menu((('start-end', 0), ('start-size', 1), ('start-end-size', 2))) \
                                    .append_visibility_condition(mode_param, '==', 'range')
            with ui.parameters_on_same_line_block():
                ui.add_parameter('range start', 'Start', NodeParameterType.FLOAT, 0).append_visibility_condition(mode_param, '==', 'range')
                ui.add_parameter('range end', 'End', NodeParameterType.FLOAT, 100).append_visibility_condition(mode_param, '==', 'range')
                ui.add_parameter('range inc', 'Inc', NodeParameterType.FLOAT, 1).append_visibility_condition(mode_param, '==', 'range')
            split_by_param = ui.add_parameter('range split by', 'Split By', NodeParameterType.INT, 0) \
                                .add_menu((('chunk size', 0), ('number of chunks', 1))) \
                                .append_visibility_condition(mode_param, '==', 'range')
            ui.add_parameter('range chunk', 'Chunk Size', NodeParameterType.INT, 10) \
                .append_visibility_condition(mode_param, '==', 'range') \
                .append_visibility_condition(split_by_param, '==', 0)
            ui.add_parameter('range count', 'Chunks Count', NodeParameterType.INT, 10) \
                .append_visibility_condition(mode_param, '==', 'range') \
                .append_visibility_condition(split_by_param, '==', 1)

            ui.add_parameter('out range type', 'Type', NodeParameterType.INT, 0) \
                .add_menu((('Int', 0), ('Float', 1))) \
                .append_visibility_condition(mode_param, '==', 'range')

            ui.add_separator()
            ui.add_parameter('out start name', 'Output Range Start Attribute Name', NodeParameterType.STRING, 'chunk_start') \
                .append_visibility_condition(mode_param, '==', 'range')
            ui.add_parameter('out end name', 'Output Range End Attribute Name', NodeParameterType.STRING, 'chunk_end') \
                .append_visibility_condition(mode_param, '==', 'range') \
                .append_visibility_condition(range_mode_param, 'in', (0, 2))
            ui.add_parameter('out size name', 'Output Range Size Attribute Name', NodeParameterType.STRING, 'chunk_size') \
                .append_visibility_condition(mode_param, '==', 'range') \
                .append_visibility_condition(range_mode_param, 'in', (1, 2))

    @staticmethod
    def _ranges_from_chunksize(start, end, inc, chunk_size):
        if chunk_size <= 0:
            raise ProcessingError('chunk size must be positive')
        ranges = []
        cur = start
        while cur < end:
            count_till_end = math.ceil((end - cur) / inc)
            if count_till_end == 0:
                break
            ranges.append((cur, min(end, cur + inc * min(count_till_end, chunk_size)), min(count_till_end, chunk_size)))
            cur = ranges[-1][1]
        return ranges

    @staticmethod
    def _ranges_from_chunkcount(start, end, inc, chunk_count):
        if chunk_count <= 0:
            raise ProcessingError('chunk count must be positive')
        ranges = []
        print(f'- {start}, {end} :{inc}')
        for i in range(chunk_count):
            e1 = start + math.ceil(((end - start) / chunk_count * i) / inc) * inc
            e2 = min(end, start + math.ceil(((end - start) / chunk_count * (i + 1)) / inc) * inc)
            print(e1, e2)
            if e1 >= e2:  # case where chunk_count is bigger than possible
                continue
            new_range = (e1, e2, math.ceil((e2-e1)/inc))
            ranges.append(new_range)
        return ranges

    def process_task(self, context):
        mode = context.param_value('split type')

        if mode == 'list':
            attrs = context.task_attributes()
            attr_name = context.param_value('attribute name')
            if attr_name not in attrs:
                raise ProcessingError('attribute "frames" not found')
            attr_value = attrs[attr_name]
            if not isinstance(attr_value, list):
                raise ProcessingError(f'attribute "{attr_name}" must be a list')
            res = ProcessingResult()
            chunksize = context.param_value('chunk size')

            split_into = 1 + (len(attr_value) - 1) // chunksize
            # yes we can split into 1 part. this should behave the same way as when splitting into multiple parts
            # in order to have consistent behaviour in the graph
            res.split_task(split_into)
            for i in range(split_into):
                res.set_split_task_attrib(i, attr_name, attr_value[chunksize*i:chunksize*(i+1)])
            return res
        elif mode == 'range':
            res = ProcessingResult()
            start = context.param_value('range start')
            end = context.param_value('range end')
            inc = context.param_value('range inc')

            rtype = context.param_value('out range type')
            split_by_type = context.param_value('range split by')
            if rtype == 0:  # int
                start = int(start)
                end = int(end)
                inc = int(inc)

            if inc <= 0:
                raise ProcessingError('inc must be positive')
            if end <= start:  # not equal cuz end specifies NEXT element after last
                raise ProcessingError('end cannot be less or equal than start')

            if split_by_type == 0:  # chunk size
                chunk_size = context.param_value('range chunk')
                ranges = self._ranges_from_chunksize(start, end, inc, chunk_size)
            elif split_by_type == 1:  # number of chunks
                chunk_count = context.param_value('range count')
                ranges = self._ranges_from_chunkcount(start, end, inc, chunk_count)
            else:
                raise NotImplementedError('wtf is this split-by mode?')

            rstart_name = context.param_value('out start name')
            rend_name = context.param_value('out end name')
            rsize_name = context.param_value('out size name')

            rmode = context.param_value('range mode')
            res.split_task(len(ranges))
            for i, (start, end, count) in enumerate(ranges):
                res.set_split_task_attrib(i, rstart_name, start)
                if rmode in (0, 2):
                    res.set_split_task_attrib(i, rend_name, end)
                if rmode in (1, 2):
                    res.set_split_task_attrib(i, rsize_name, count)
            return res
        else:
            raise NotImplementedError(f'mode "{mode}" is not implemented')


if __name__ == '__main__':
    # some fast tests
    r = FramerangeSplitter._ranges_from_chunkcount(10, 12, 3, 10)
    assert r == [(10, 12, 1)], r
    r = FramerangeSplitter._ranges_from_chunkcount(10, 15, 3, 10)
    assert r == [(10, 13, 1), (13, 15, 1)], r
    r = FramerangeSplitter._ranges_from_chunkcount(15, 16, 3, 10)
    assert r == [(15, 16, 1)], r
    r = FramerangeSplitter._ranges_from_chunkcount(15, 29, 3, 3)
    assert r == [(15, 21, 2), (21, 27, 2), (27, 29, 1)], r
    r = FramerangeSplitter._ranges_from_chunkcount(15, 29, 3, 4)
    assert r == [(15, 21, 2), (21, 24, 1), (24, 27, 1), (27, 29, 1)], r

    r = FramerangeSplitter._ranges_from_chunksize(10, 12, 3, 1)
    assert r == [(10, 12, 1)], r
    r = FramerangeSplitter._ranges_from_chunksize(10, 12, 3, 10)
    assert r == [(10, 12, 1)], r
    r = FramerangeSplitter._ranges_from_chunksize(10, 15, 3, 1)
    assert r == [(10, 13, 1), (13, 15, 1)], r
    r = FramerangeSplitter._ranges_from_chunksize(10, 15, 3, 10)
    assert r == [(10, 15, 2)], r
    r = FramerangeSplitter._ranges_from_chunksize(15, 16, 3, 10)
    assert r == [(15, 16, 1)], r
    r = FramerangeSplitter._ranges_from_chunksize(15, 38, 3, 3)
    assert r == [(15, 24, 3), (24, 33, 3), (33, 38, 2)], r



from copy import copy
from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern

from typing import Iterable


def node_class():
    return HipAssGenerator


class HipAssGenerator(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'ass generator'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'hip', 'houdini', 'ass', 'arnold', 'generator', 'render', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'hip_ass_generator'

    def __init__(self, name):
        super(HipAssGenerator, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.0, 0.54, 0.56)
            ui.add_output_for_spawned_tasks()
            ui.add_parameter('hip path', 'hip file path', NodeParameterType.STRING, "`task['hipfile']`")
            with ui.parameters_on_same_line_block():
                mask_hip = ui.add_parameter('mask as different hip', 'mask as different hip file', NodeParameterType.BOOL, False)
                ui.add_parameter('mask hip path', '', NodeParameterType.STRING, "`task.get('hipfile_orig', task['hipfile'])`")\
                    .append_visibility_condition(mask_hip, '==', True)
            ui.add_parameter('driver path', 'mantra node path', NodeParameterType.STRING, "`task['hipdriver']`")
            ui.add_parameter('attrs to context', 'set these attribs as context', NodeParameterType.STRING, '')
            ui.add_parameter('ass file path', 'ass file path', NodeParameterType.STRING, "`config['global_scratch_location']`/`node.name`/`task.name`_`task.id`/ass/`node.name`.$F4.ass")
            with ui.parameters_on_same_line_block():
                skipparam = ui.add_parameter('skip if exists', 'skip if result already exists', NodeParameterType.BOOL, False)
                ui.add_parameter('gen for skipped', 'generate children for skipped', NodeParameterType.BOOL, True).append_visibility_condition(skipparam, '==', True)
            ui.add_parameter('attrs', 'attributes to copy to children', NodeParameterType.STRING, '')

            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    def process_task(self, context) -> ProcessingResult:
        """
        this node expects to find the following attributes:
        frames
        hipfile
        hipdriver
        :param context:
        :return:
        """
        attrs = context.task_attributes()
        if any(x not in attrs for x in ('frames',)):
            raise ProcessingError('required attribute "frames" not found')
        hippath = context.param_value('hip path')
        driverpath = context.param_value('driver path')
        frames = attrs['frames']
        matching_attrnames = filter_by_pattern(context.param_value('attrs'), attrs.keys())
        attr_to_trans = tuple((x, attrs[x]) for x in matching_attrnames if x not in ('frames', 'file'))
        attr_to_context = filter_by_pattern(context.param_value('attrs to context'), attrs.keys())

        env = InvocationEnvironment()

        spawnlines = \
            f"        filepath = node.evalParm('ar_ass_file')\n" \
            f"        outimage = node.evalParm('ar_picture')\n" \
            f"        attrs = {{'frames': [frame], 'file': filepath, 'hipfile': {repr(hippath)}, 'outimage': outimage}}\n" \
            f"        for attr, val in {repr(attr_to_trans)}:\n" \
            f"            attrs[attr] = val\n" \
            f"        lifeblood_connection.create_task(node.name() + '_spawned frame %g' % frame, attrs)\n"

        if not self.is_output_connected('spawned'):
            spawnlines = ''

        script = \
            f'import os\n' \
            f'import hou\n' \
            f'import lifeblood_connection\n'

        if context.param_value('mask as different hip'):
            mask_path = context.param_value('mask hip path')
            script += 'def __fix_hip_env__(event_type=None):\n' \
                      '    if event_type == hou.hipFileEventType.BeforeSave:\n' \
                      '        hou.hipFile.setName(os.devnull)\n' \
                      '        return\n' \
                      '    if event_type not in (hou.hipFileEventType.AfterSave, hou.hipFileEventType.AfterLoad) and event_type is not None:\n' \
                      '        return\n' \
                     f'    hou.hipFile.setName({repr(mask_path)})\n' \
                      'hou.hipFile.addEventCallback(__fix_hip_env__)\n'

        script += \
            f'print("opening file" + {repr(hippath)})\n' \
            f'hou.hipFile.load({repr(hippath)}, ignore_load_warnings=True)\n' \
            f'node = hou.node({repr(driverpath)})\n'

        for attrname in attr_to_context:
            script += f'hou.setContextOption({repr(attrname)}, {repr(attrs[attrname])})\n'

        script += \
            f'if node.parm("ar_ass_export_enable").evalAsInt() != 1:\n' \
            f'    node.parm("ar_ass_export_enable").set(1)\n'
        asspath = context.param_value('ass file path').strip()
        if asspath != '':
            script += \
                f'node.parm("ar_ass_file").set({repr(asspath)})\n'
        script += \
            f'for frame in {repr(frames)}:\n' \
            f'    hou.setFrame(frame)\n' \
            f'    skipped = False\n'
        if context.param_value('skip if exists'):
            script += \
                f'    skipped = os.path.exists(node.parm("ar_ass_file").evalAsString())\n'
        script += \
            f'    if skipped:\n' \
            f'        print("output file already exists, skipping frame %d" % frame)\n' \
            f'    else:\n' \
            f'        print("rendering frame %d" % frame)\n' \
            f'        node.render(frame_range=(frame, frame))\n'
        if spawnlines is not None:
            script += \
                f'    if {repr(context.param_value("gen for skipped"))} or not skipped:\n' \
                f'{spawnlines}'
        script += \
            f'print("all done!")\n'

        inv = InvocationJob(['hython', ':/work_to_do.py'], env=env)
        inv.set_extra_file('work_to_do.py', script)
        res = ProcessingResult(job=inv)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import uuid
import psutil
import pickle
import copy
import re
from .base import TypeMetadata
from .misc import get_unique_machine_id
from .logging import get_logger

from typing import Optional, TYPE_CHECKING, Tuple, Type, Union, Set
if TYPE_CHECKING:
    from .basenode import BaseNode

__logger = get_logger('worker_resources')


__mem_parse_re = re.compile(r'^\s*(\d+(?:\.\d+)?)\s*([BKMGTP]?)\s*$')


def _try_parse_mem_spec(s: Union[str, int], default: Optional[int] = None):
    if not isinstance(s, str):
        return s
    match = __mem_parse_re.match(s)
    if not match:
        __logger.warning(f'could not parse "{s}", using default')
        return default
    bytes_count = float(match.group(1))
    coeff = match.group(2)

    coeff_map = {'B': 1,
                 'K': 10**3,
                 'M': 10**6,
                 'G': 10**9,
                 'T': 10**12,
                 'P': 10**15}

    if coeff not in coeff_map:
        __logger.warning(f'could not parse "{s}", wtf is "{coeff}"? using default')
        return default

    if coeff:
        mult = coeff_map[coeff]

        # this way we limit float op errors
        if mult > 10**6:
            mult //= 10**6
            bytes_count = int(bytes_count * 10**6)

        bytes_count = bytes_count * mult

    return int(bytes_count)


class NodeTypeMetadata(TypeMetadata):
    def __init__(self, node_type: Type["BaseNode"]):
        from . import pluginloader  # here cuz it should only be created from lifeblood, but can be used from viewer too
        self.__type_name = node_type.type_name()
        self.__label = node_type.label()
        self.__tags = set(node_type.tags())
        self.__description = node_type.description()
        self.__settings_names = tuple(pluginloader.nodes_settings.get(node_type.type_name(), {}).keys())

    @property
    def type_name(self) -> str:
        return self.__type_name

    @property
    def label(self) -> Optional[str]:
        return self.__label

    @property
    def tags(self) -> Set[str]:
        return self.__tags

    @property
    def description(self) -> str:
        return self.__description

    @property
    def settings_names(self) -> Tuple[str, ...]:
        return self.__settings_names


class WorkerResources:
    __res_names = ('cpu_count', 'cpu_mem', 'gpu_count', 'gpu_mem')  # name of all main resources
    __resource_epsilon = 1e-5

    def __init__(self,
                 cpu_count: Union[int, float, None] = None, cpu_mem: Optional[int] = None,
                 gpu_count: Union[int, float, None] = None, gpu_mem: Optional[int] = None):
        """

        NOTE: because of float resource rounding - it's not safe to rely on consistency of summing/subtracting a lot of resources
        """
        self.hwid = get_unique_machine_id()
        self.cpu_count = cpu_count if cpu_count is not None else psutil.cpu_count()  # note, cpu/gpu count can be float by design, but we don't like "almost" round float values
        self.cpu_mem = _try_parse_mem_spec(cpu_mem) if cpu_mem is not None else psutil.virtual_memory().total
        self.gpu_count = gpu_count if gpu_count is not None else 0  # TODO: implement GPU autodetection
        self.gpu_mem = _try_parse_mem_spec(gpu_mem) if gpu_mem is not None else 0
        self.total_cpu_count = self.cpu_count
        self.total_cpu_mem = self.cpu_mem
        self.total_gpu_count = self.gpu_count
        self.total_gpu_mem = self.gpu_mem

        for rname in self.__res_names:  # sanity check
            assert hasattr(self, rname)
            assert hasattr(self, f'total_{rname}')

    def serialize(self) -> bytes:
        return pickle.dumps(self)

    @classmethod
    def deserialize(cls, data) -> "WorkerResources":
        return pickle.loads(data)

    def is_valid(self):
        return all(getattr(self, res) >= 0 for res in self.__res_names)
        # return self.cpu_count >= 0 and \
        #        self.cpu_mem >= 0 and \
        #        self.gpu_count >= 0 and \
        #        self.gpu_mem >= 0

    def __repr__(self):
        return f'<cpu: {self.cpu_count}/{self.total_cpu_count}, mem: {self.cpu_mem}/{self.total_cpu_mem}, gpu: {self.gpu_count}/{self.total_gpu_count}, gmm: {self.gpu_mem}/{self.total_gpu_mem}>'

    def __lt__(self, other):
        if not isinstance(other, WorkerResources):
            return other > self
        return any(getattr(self, res) < getattr(other, res) for res in self.__res_names)
        # return self.cpu_count < other.cpu_count or \
        #        self.cpu_mem < other.cpu_mem or \
        #        self.gpu_count < other.gpu_count or \
        #        self.gpu_mem < other.gpu_mem

    def __eq__(self, other):
        """
        note - this does NOT compare totals, only actual resources
        :param other:
        :return:
        """
        if not isinstance(other, WorkerResources):
            return other == self
        return all(getattr(self, res) == getattr(other, res) for res in self.__res_names)
        # return self.cpu_count == other.cpu_count and \
        #        self.cpu_mem == other.cpu_mem and \
        #        self.gpu_count == other.gpu_count and \
        #        self.gpu_mem == other.gpu_mem

    def __ne__(self, other):
        return not (self == other)

    def __le__(self, other):
        return self < other or self == other

    def __sub__(self, other):
        res = copy.copy(self)
        for resname in self.__res_names:
            val = getattr(res, resname) - getattr(other, resname)
            if isinstance(val, float):
                rounded_val = round(val)
                if abs(val - rounded_val) <= self.__resource_epsilon:
                    val = rounded_val
            setattr(res, resname, val)
        # res.cpu_count -= other.cpu_count
        # res.cpu_mem -= other.cpu_mem
        # res.gpu_count -= other.gpu_count
        # res.gpu_mem -= other.gpu_mem
        return res

    def __add__(self, other):
        res = copy.copy(self)
        for resname in self.__res_names:
            val = getattr(res, resname) - getattr(other, resname)
            if isinstance(val, float):
                rounded_val = round(val)
                if abs(val - rounded_val) <= self.__resource_epsilon:
                    val = rounded_val
            setattr(res, resname, val)
        # res.cpu_count += other.cpu_count
        # res.cpu_mem += other.cpu_mem
        # res.gpu_count += other.gpu_count
        # res.gpu_mem += other.gpu_mem
        return res





import sys
import os
from lifeblood.basenode import BaseNodeWithTaskRequirements, ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.enums import NodeParameterType
from lifeblood.invocationjob import InvocationJob
from lifeblood.paths import config_path
from typing import Iterable
import subprocess
import time
import json
import tempfile
import shutil


description = \
'''
'''


def node_class():
    return SimpleWebServer


class SimpleWebServer(BaseNodeWithTaskRequirements):
    def __init__(self, name):
        super().__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.051, 0.741, 0.545)
            ui.add_output_for_spawned_tasks()
            with ui.parameters_on_same_line_block():
                ui.add_parameter('address', 'Listen Address', NodeParameterType.STRING, '127.0.0.1')
                ui.add_parameter('port', 'Port', NodeParameterType.INT, 8080)
            ui.add_separator()
            ui.add_parameter('expect_ready_message', 'Expect "Processing Ready" Message', NodeParameterType.BOOL, False)
            ui.add_separator()
            ui.add_parameter('page_title', 'Page Message', NodeParameterType.STRING, '<strong>File upload</strong> example')
            ui.add_parameter('base_path', 'Base Dir For Input Files', NodeParameterType.STRING, f"`config['global_scratch_location']`/{self.type_name()}/input")

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        port = context.param_value('port')
        addr = context.param_value('address')
        title = context.param_value('page_title')
        base_path = context.param_value('base_path')
        expect_reply = context.param_value('expect_ready_message')

        job = InvocationJob(['python', self.my_plugin().package_data() / "server.py",
                             '--ip', addr,
                             '--port', port,
                             '--title', title,
                             '--base_storage_path', base_path] +
                            (['--expect-reply'] if expect_reply else []))

        return ProcessingResult(job)

    @classmethod
    def type_name(cls) -> str:
        return 'stock_simple_web_server'

    @classmethod
    def label(cls) -> str:
        return 'simple web server'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'stock', 'web', 'server'

    @classmethod
    def description(cls) -> str:
        return description



import asyncio
from pathlib import Path
import sqlite3
from warnings import warn
from aiosqlite import *
from .logging import get_logger

from typing import Any, Callable, List, Optional, Tuple, Union


logger = get_logger('aiosqlite_overlay')


class ConnectionWithCallbacks(Connection):
    def __init__(self, *args, pragmas_after_connect=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.__callbacks: List[Tuple[Callable, tuple, dict]] = []
        self.__post_connect_pragmas = pragmas_after_connect

    def add_after_commit_callback(self, callable: Callable, *args, **kwargs):
        self.__callbacks.append((callable, args, kwargs))

    async def __aenter__(self) -> Connection:
        con = await super().__aenter__()
        if self.__post_connect_pragmas is not None:
            for statement in self.__post_connect_pragmas:
                await con.execute(f'PRAGMA {statement}')
        return con

    async def commit(self):
        await super().commit()
        for callback, args, kwargs in self.__callbacks:
            try:
                callback(*args, **kwargs)
            except Exception as e:
                logger.exception(f'failed to call post-commit callback {e}')
        self.__callbacks = []

    async def rollback(self) -> None:
        await super().rollback()
        self.__callbacks = []

#
# the function below is a direct copy from aiosqlite with the change of Connection class
# PLEASE don't forget to check this is asyncio version changes much
def connect(
    database: Union[str, Path],
    *,
    iter_chunk_size=64,
    loop: Optional[asyncio.AbstractEventLoop] = None,
    pragmas_after_connect=None,
    **kwargs: Any
) -> ConnectionWithCallbacks:
    """Create and return a connection proxy to the sqlite database."""

    if loop is not None:
        warn(
            "aiosqlite.connect() no longer uses the `loop` parameter",
            DeprecationWarning,
        )

    def connector() -> sqlite3.Connection:
        if isinstance(database, str):
            loc = database
        elif isinstance(database, bytes):
            loc = database.decode("utf-8")
        else:
            loc = str(database)

        return sqlite3.connect(loc, **kwargs)

    return ConnectionWithCallbacks(connector, iter_chunk_size, pragmas_after_connect=pragmas_after_connect)



from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern

from typing import Iterable


def node_class():
    return BlenderBatchRender


class BlenderBatchRender(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'blender batch render'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'blender', 'batch', 'render', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'blender_batch_render'

    @classmethod
    def description(cls) -> str:
        return 'This node renders a blender file.\n' \
               '\n' \
               '- If second input is disconnected: attribute "images" will be created\n' \
               '  on processed task containing a list of all images rendered by it.\n' \
               '- If second input is connected: attribute "images" will NOT be created\n' \
               '  on processed task, instead a child task will be spawned from the second\n' \
               '  input for EACH frame with "file" attribute having the path to the image\n' \
               '  Also spawned children tasks will get all attributes from the parent that\n' \
               '  match "attributes to copy to children" parameter mask\n' \
               '\n' \
               'Pre Script: if provided - this python script is run on the given file BEFORE render\n' \
               'If in addition "Skip Actual Render" checkbox is set - then ONLY this "Pre Script"\n' \
               'will be executed, no render will be done\n' \
               '\n' \
               '(Note: currently you CANNOT modify task attributes from the pre-script)'

    def __init__(self, name):
        super(BlenderBatchRender, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.922, 0.467, 0.0)
            ui.color_scheme().set_secondary_color(0.129, 0.337, 0.533)
            ui.add_output_for_spawned_tasks()

            ui.add_parameter('file_path', 'File Path', NodeParameterType.STRING, "`task['file']`")
            with ui.parameters_on_same_line_block():
                do_over = ui.add_parameter('do_override_frame', 'Override Frame', NodeParameterType.BOOL, False)
                ui.add_parameter('override_frame_start', None, NodeParameterType.INT, 1) \
                    .append_visibility_condition(do_over, '==', True)
                ui.add_parameter('override_frame_end', None, NodeParameterType.INT, 100) \
                    .append_visibility_condition(do_over, '==', True)
                ui.add_parameter('override_frame_step', None, NodeParameterType.INT, 1) \
                    .append_visibility_condition(do_over, '==', True)
                ui.add_parameter('frames_attrib_name', 'Frames Attribute', NodeParameterType.STRING, 'frames') \
                    .append_visibility_condition(do_over, '==', False)
            with ui.parameters_on_same_line_block():
                do_over = ui.add_parameter('do_override_scene', 'Override Scene', NodeParameterType.BOOL, False)
                ui.add_parameter('override_scene', None, NodeParameterType.STRING, 'scene001') \
                    .append_visibility_condition(do_over, '==', True)
            with ui.parameters_on_same_line_block():
                do_over = ui.add_parameter('do_override_output', 'Override Output', NodeParameterType.BOOL, False)
                ui.add_parameter('override_output', None, NodeParameterType.STRING, '//something.####.png') \
                    .append_visibility_condition(do_over, '==', True)
            ui.add_parameter('attrs', 'attributes to copy to children', NodeParameterType.STRING, '*')

            ui.add_separator()

            do_over = ui.add_parameter('do_pre_script', 'Pre Script', NodeParameterType.BOOL, False)
            ui.add_parameter('pre_script', None, NodeParameterType.STRING, '# task var is available here\n# get task attribute with task[\'attr_name\']\n\n') \
                .append_visibility_condition(do_over, '==', True) \
                .set_text_multiline(syntax_hint='python')
            ui.add_parameter('skip_render', 'Skip Actual Render', NodeParameterType.BOOL, False) \
                .append_visibility_condition(do_over, '==', True)

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        file_path = context.param_value('file_path')
        do_frames = context.param_value('do_override_frame')
        do_scene = context.param_value('do_override_scene')
        do_output = context.param_value('do_override_output')
        do_prescript = context.param_value('do_pre_script')

        script_prefix = f'class _TaskWrapper:\n' \
                        f'    def __init__(self):\n' \
                        f'        self.__attrs = {repr(dict(context.task_attributes()))}\n' \
                        f'    @property\n' \
                        f'    def name(self):\n' \
                        f'        return {repr(context.task_field("name"))}\n' \
                        f'    @property\n' \
                        f'    def id(self):\n' \
                        f'        return {repr(context.task_id())}\n' \
                        f'    def attrib_names(self):\n' \
                        f'        return tuple(self.__attrs.keys())\n' \
                        f'    def __getitem__(self, key):\n' \
                        f'        return self.__attrs[key]\n' \
                        f'\n' \
                        f'task = _TaskWrapper()\n' \
                        f'\n\n'
        scripts = {}

        args = ['blender', '-b', file_path]

        if do_scene:
            args += ['-S', context.param_value('override_scene')]

        if do_output:
            args += ['-o', context.param_value('override_output')]

        if do_prescript:
            scripts['prescript.py'] = script_prefix + context.param_value('pre_script')
            args += ['-P', ':/prescript.py']

        # if not self.is_output_connected('spawned'):  # in case we spawn tasks we need to do everything very different
        #     if do_frames:
        #         f_start = context.param_value('override_frame_start')
        #         f_end = context.param_value('override_frame_end')
        #         f_step = context.param_value('override_frame_step')
        #         args += ['-j', f'{f_step}']
        #         args += ['-f', f'{f_start}..{f_end}']
        #     else:
        #         frames_name = context.param_value('frames_attrib_name')
        #         if not context.task_has_attribute(frames_name):
        #             raise ProcessingError(f'attribute "{frames_name}" not found')
        #         frames = context.task_attribute(frames_name)
        #         if not isinstance(frames, list):
        #             if not isinstance(frames, (int, float)):
        #                 raise ProcessingError(f'attribute "{frames_name}" is of wrong type "{type(frames)}"')
        #             frames = [frames]
        #         if any(not isinstance(x, (int, float)) for x in frames):
        #             raise ProcessingError(f'attribute "{frames_name}" has non numeric items')
        #         args += ['-f', ','.join(str(x) for x in frames)]
        #
        # else:
        attrs = context.task_attributes()
        matching_attrnames = filter_by_pattern(context.param_value('attrs'), attrs.keys())
        attr_to_trans = tuple((x, attrs[x]) for x in matching_attrnames if x not in ('frames', 'file', 'outimage'))

        if do_frames:
            f_start = context.param_value('override_frame_start')
            f_end = context.param_value('override_frame_end')
            f_step = context.param_value('override_frame_step')
            # currently override frames parameters are INTs, so we can do range
            frames = list(range(f_start, f_end+1, f_step))  # end inclusive
        else:
            frames_name = context.param_value('frames_attrib_name')
            if not context.task_has_attribute(frames_name):
                raise ProcessingError(f'attribute "{frames_name}" not found')
            frames = context.task_attribute(frames_name)
            if not isinstance(frames, list):
                if not isinstance(frames, (int, float)):
                    raise ProcessingError(f'attribute "{frames_name}" is of wrong type "{type(frames)}"')
                frames = [frames]
            if any(not isinstance(x, (int, float)) for x in frames):
                raise ProcessingError(f'attribute "{frames_name}" has non numeric items')

        # is correct one frame_current or frame_current_final ?
        spawnlines = \
            f"    attrs = {{'frames': [frame], 'blendfile': {repr(file_path)}, 'file': outimage}}\n" \
            f"    for attr, val in {repr(attr_to_trans)}:\n" \
            f"        attrs[attr] = val\n" \
            f"    lifeblood_connection.create_task('{context.task_name()}_spawned frame %g' % frame, attrs)\n"

        do_spawn = self.is_output_connected('spawned')
        main_script = f'import bpy\n' \
                      f'import lifeblood_connection\n' \
                      f'scene = bpy.context.scene\n' \
                      f'all_outimages = []\n' \
                      f'for frame in {repr(frames)}:\n' \
                      f'    scene.frame_set(int(frame), subframe=frame%1.0)\n' \
                      f'    outimage = scene.render.frame_path()\n' \
                      f'    all_outimages.append(outimage)\n' \
                      f'    _filepath_stash = scene.render.filepath\n' \
                      f'    scene.render.filepath = outimage\n' \
                      f'    bpy.ops.render.render(write_still=True)\n' \
                      + (spawnlines if do_spawn else '') + \
                      f'    scene.render.filepath = _filepath_stash\n' \
                      + ('lifeblood_connection.set_attributes({"images": all_outimages})\n' if not do_spawn else '') \
                      + (f'lifeblood_connection.wait_for_currently_running_async_operations()\n' if do_spawn else '')

        if not context.param_value('do_pre_script') or not context.param_value('skip_render'):
            scripts['main_script.py'] = main_script
            args += ['-P', ':/main_script.py']

        invoc = InvocationJob(args)
        for file_name, file_contents in scripts.items():
            invoc.set_extra_file(file_name, file_contents)
        return ProcessingResult(invoc)



import asyncio
import struct
import logging
from lifeblood.logging import get_logger
from datetime import datetime
from dataclasses import dataclass
from ..exceptions import MessageTransferError, MessageTransferTimeoutError
from ..interfaces import MessageStreamFactory
from ..stream_wrappers import MessageSendStream, MessageSendStreamBase
from ..address import DirectAddress, AddressChain
from ..defaults import default_stream_timeout
from ..messages import Message

from typing import Awaitable, Callable, Dict, List, Optional, Tuple


async def _initialize_connection(destination: DirectAddress, source: DirectAddress) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
    host, sport = destination.split(':')
    reader, writer = await asyncio.open_connection(host, int(sport))
    source_bytes = source.encode('UTF-8')
    writer.write(struct.pack('>Q', len(source_bytes)))
    writer.write(source_bytes)
    await writer.drain()
    return reader, writer


@dataclass
class ConnectionPoolEntry:
    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter
    last_used: datetime
    users_count: int
    close_when_user_count_zero: bool = False
    bad: bool = False


class ReusableMessageSendStream(MessageSendStream):
    """

    """
    def __init__(self,
                 pool_entry: ConnectionPoolEntry,
                 *,
                 reply_address: DirectAddress,
                 destination_address: DirectAddress,
                 stream_timeout: float = default_stream_timeout,
                 confirmation_timeout: float = default_stream_timeout):
        super().__init__(pool_entry.reader,
                         pool_entry.writer,
                         reply_address=reply_address,
                         destination_address=destination_address,
                         stream_timeout=stream_timeout,
                         confirmation_timeout=confirmation_timeout)
        self.__pool_entry = pool_entry
        self.__closed = False
        self.__pool_entry.users_count += 1
        self.__do_actually_wait_closed = False

    async def send_raw_message(self, message: Message, *, message_delivery_timeout_override: Optional[float] = ...):
        try:
            return await super().send_raw_message(message, message_delivery_timeout_override=message_delivery_timeout_override)
        except MessageTransferTimeoutError:
            # we cannot be sure some crap won't arrive after timeout,
            # in that case future uses of this connection will be at risk of getting it,
            # so it's safer to mark it for closure
            self.__pool_entry.close_when_user_count_zero = True
            raise

    def close(self):
        if self.__closed:
            return
        self.__closed = True
        # note: section below might need locking when close() can be called from different threads
        #  actually, there is much more problems if streams are opened/closed from different threads
        self.__pool_entry.last_used = datetime.now()
        self.__pool_entry.users_count -= 1
        if self.__pool_entry.users_count == 0 and self.__pool_entry.close_when_user_count_zero:
            super().close()
            self.__do_actually_wait_closed = True

    async def wait_closed(self):
        if not self.__do_actually_wait_closed:
            return
        await super().wait_closed()

    def __del__(self):
        self.close()  # multiple close() calls are fine


class TcpMessageStreamPooledFactory(MessageStreamFactory):
    _logger: Optional[logging.Logger] = None

    def __init__(self,
                 pooled_connection_life: int = 0,
                 connection_open_function: Optional[Callable[[DirectAddress, DirectAddress], Awaitable[Tuple[asyncio.StreamReader, asyncio.StreamWriter]]]] = None,
                 timeout: float = default_stream_timeout):
        self.__pooled_connection_life = pooled_connection_life
        self.__pool: Dict[Tuple[str, int], List[ConnectionPoolEntry]] = {}
        self.__connection_open_func: Callable[[DirectAddress, DirectAddress], Awaitable[Tuple[asyncio.StreamReader, asyncio.StreamWriter]]] = connection_open_function or _initialize_connection
        self.__open_connection_calls_count = 0
        self.__pool_closed = asyncio.Event()
        self.__timeout = timeout
        if self._logger is None:
            TcpMessageStreamPooledFactory._logger = get_logger('TcpMessageStreamPooledFactory')

        self.__closing_task_to_wait = None

    async def prune(self):
        await self.close_unused_connections_older_than(self.__pooled_connection_life)

    async def close_unused_connections_older_than(self, older_than_this_seconds: float):
        keys_to_remove = []
        now = datetime.now()
        to_wait = []
        for key, entry_list in self.__pool.items():
            new_entries = []
            for entry in entry_list:
                if not entry.bad \
                        and (entry.users_count > 0
                             or ((now - entry.last_used).total_seconds() < older_than_this_seconds
                                 and not entry.close_when_user_count_zero)
                             ):
                    new_entries.append(entry)
                else:
                    if not entry.writer.is_closing():
                        entry.writer.close()
                    to_wait.append(asyncio.create_task(entry.writer.wait_closed()))
            if len(new_entries) == 0:
                keys_to_remove.append(key)
            else:
                self.__pool[key] = new_entries
        for key in keys_to_remove:
            self.__pool.pop(key)

        if to_wait:
            dones, _ = await asyncio.wait(to_wait, return_when=asyncio.ALL_COMPLETED)
            for task in dones:
                exc = task.exception()
                if exc is None:
                    continue
                self._logger.warning(f'pooled connection was faulty and failed to close: "{exc}"')

    def _get_cached_entry(self, host: str, port: int) -> Optional[ConnectionPoolEntry]:
        key = (host, port)
        if key not in self.__pool:
            return None

        entry_list = self.__pool[key]
        # to_remove = []
        selected: Optional[ConnectionPoolEntry] = None
        for entry in entry_list:
            if entry.users_count > 0 or entry.close_when_user_count_zero:
                continue
            if entry.bad or entry.reader.at_eof() or entry.writer.is_closing():
                # to_remove.append(entry)
                entry.bad = True
                continue
            selected = entry
        # for entry in to_remove:  # not the most optimal way......
        #     entry_list.remove(entry)
        return selected

    async def open_sending_stream(self, destination: DirectAddress, source: DirectAddress) -> MessageSendStreamBase:
        if self.__pool_closed.is_set():
            raise RuntimeError('cannot open sending stream when pool closed')
        host, sport = destination.split(':')
        port = int(sport)
        entry = self._get_cached_entry(host, port)
        if entry is not None:  # test entry with ping message
            stream = ReusableMessageSendStream(entry,
                                               reply_address=source,
                                               destination_address=destination,
                                               stream_timeout=self.__timeout,
                                               confirmation_timeout=self.__timeout)
            try:
                await stream.send_ping()
            except MessageTransferError as e:
                self._logger.debug('ping failed due to %s', e)
                entry.bad = True
                entry = None
                stream.close()

        if entry is None:
            key = (host, port)
            reader, writer = await self.__connection_open_func(destination, source)
            self.__open_connection_calls_count += 1
            entry = ConnectionPoolEntry(reader,
                                        writer,
                                        datetime.now(),
                                        0)
            self.__pool.setdefault(key, []).append(entry)
        assert entry is not None

        # Warning: be sure there is no possibility of prune running (no awaits)
        #  between here (where use_count incs)
        #  and where it was fetched/created. We don't want to lose connection we've just got
        stream = ReusableMessageSendStream(entry,
                                           reply_address=source,
                                           destination_address=destination,
                                           stream_timeout=self.__timeout,
                                           confirmation_timeout=self.__timeout)
        await self.prune()
        return stream

    async def close_all_unused_connections(self):
        await self.close_unused_connections_older_than(-1)

    def close_pool(self):
        self.__closing_task_to_wait = asyncio.create_task(self.__closing_task())
        self.__pool_closed.set()
        for items in self.__pool.values():
            for item in items:
                item.close_when_user_count_zero = True

    async def wait_pool_closed(self):
        await self.__pool_closed.wait()
        # by this time self.__closing_task_to_wait will be not None
        await self.__closing_task_to_wait

    async def __closing_task(self):
        await self.__pool_closed.wait()
        while len(self.__pool) > 0:
            await self.close_all_unused_connections()
            await asyncio.sleep(0.5)  # too heavy polling?

    def __del__(self):
        self._logger.debug(f'connections opened total: {self.__open_connection_calls_count}')
        if len(self.__pool) > 0:
            self._logger.warning(f'pooled connections remain ({len(self.__pool)}) after __del__, {len([1 for l in self.__pool.values() for x in l if x.users_count > 0])} still in use')


class TcpMessageStreamFactory(MessageStreamFactory):
    def __init__(self, *, timeout: float = default_stream_timeout):
        self.__timeout = timeout

    async def open_sending_stream(self, destination: DirectAddress, source: DirectAddress) -> MessageSendStreamBase:
        """
        address is expected to be in form of "1.2.3.4:1313|2.3.4.5:2424"...
        """
        reader, writer = await _initialize_connection(destination, source)

        return MessageSendStream(reader, writer,
                                 reply_address=source,
                                 destination_address=destination,
                                 stream_timeout=self.__timeout,
                                 confirmation_timeout=self.__timeout)



import sys
import os
from pathlib import Path
import asyncio
import signal
from ..defaults import scheduler_port as default_scheduler_port, ui_port as default_ui_port
from ..config import get_config, create_default_user_config_file, get_local_scratch_path
from .. import logging
from .. import paths

from .scheduler import Scheduler

from typing import Optional


default_config = f'''
[core]
## you can uncomment stuff below to specify some static values
## 
# server_ip = "192.168.0.2"
# server_port = {default_scheduler_port()}
# ui_ip = "192.168.0.2"
# ui_port = {default_ui_port()}

## you can turn off scheduler broadcasting if you want to manually configure viewer and workers to connect
## to a specific address
# broadcast = false

[scheduler]

[scheduler.globals]
## entries from this section will be available to any node from config[key] 
##
## if you use more than 1 machine - you must change this to a network location shared among all workers
## by default it's set to scheduler's machine local temp path, and will only work for 1 machine setup 
global_scratch_location = "{get_local_scratch_path()}"

[scheduler.database]
## you can specify default database path, 
##  but this can be overriden with command line argument --db-path
# path = "/path/to/database.db"

## uncomment line below to store task logs outside of the database
##  it works in a way that all NEW logs will be saved according to settings below
##  existing logs will be kept where they are
##  external logs will ALWAYS be looked for in location specified by store_logs_externally_location
##  so if you have ANY logs saved externally - you must keep store_logs_externally_location defined in the config, 
##    or those logs will be inaccessible
##  but you can safely move logs and change location in config accordingly, but be sure scheduler is not accessing them at that time
# store_logs_externally = true
# store_logs_externally_location = /path/to/dir/where/to/store/logs
'''


async def main_async(db_path=None, *, broadcast_interval: Optional[int] = None):
    def graceful_closer(*args):
        scheduler.stop()

    noasync_do_close = False
    def noasync_windows_graceful_closer_event(*args):
        nonlocal noasync_do_close
        noasync_do_close = True

    async def windows_graceful_closer():
        while not noasync_do_close:
            await asyncio.sleep(1)
        graceful_closer()

    scheduler = Scheduler(db_path,
                          do_broadcasting=broadcast_interval > 0 if broadcast_interval is not None else None,
                          broadcast_interval=broadcast_interval)
    win_signal_waiting_task = None
    try:
        asyncio.get_event_loop().add_signal_handler(signal.SIGINT, graceful_closer)
        asyncio.get_event_loop().add_signal_handler(signal.SIGTERM, graceful_closer)
    except NotImplementedError:  # solution for windows
        signal.signal(signal.SIGINT, noasync_windows_graceful_closer_event)
        signal.signal(signal.SIGBREAK, noasync_windows_graceful_closer_event)
        win_signal_waiting_task = asyncio.create_task(windows_graceful_closer())

    await scheduler.start()
    await scheduler.wait_till_stops()
    if win_signal_waiting_task is not None:
        if not win_signal_waiting_task.done():
            win_signal_waiting_task.cancel()
    logging.get_logger('scheduler').info('SCHEDULER STOPPED')


def main(argv):
    import argparse
    import tempfile

    parser = argparse.ArgumentParser('lifeblood scheduler')
    parser.add_argument('--db-path', help='path to sqlite database to use')
    parser.add_argument('--ephemeral', action='store_true', help='start with an empty one time use database, that is placed into shared memory IF POSSIBLE')
    parser.add_argument('--verbosity-pinger', help='set individual verbosity for worker pinger')
    parser.add_argument('--broadcast-interval', type=int, help='help easily override broadcasting interval (in seconds). value 0 disables broadcasting')
    opts = parser.parse_args(argv)

    # check and create default config if none
    create_default_user_config_file('scheduler', default_config)

    config = get_config('scheduler')
    if opts.db_path is not None:
        db_path = opts.db_path
    else:
        db_path = config.get_option_noasync('scheduler.database.path', str(paths.default_main_database_location()))

    global_logger = logging.get_logger('scheduler')

    fd = None
    if opts.ephemeral:
        if opts.db_path is not None:
            parser.error('only one of --db-path or --ephemeral must be provided, not both')
        # 'file:memorydb?mode=memory&cache=shared'
        # this does not work ^ cuz shared cache means that all runs on the *same connection*
        # and when there is a transaction conflict on the same connection - we get instalocked (SQLITE_LOCKED)
        # and there is no way to emulate normal DB in memory but with shared cache

        # look for shm (UNIX only)
        shm_path = Path('/dev/shm')
        lb_shm_path = None
        if shm_path.exists():
            lb_shm_path = shm_path/f'u{os.getuid()}-lifeblood'
            try:
                lb_shm_path.mkdir(exist_ok=True)
            except Exception as e:
                global_logger.warning('/dev/shm is not accessible (permission issues?), creating ephemeral database in temp dir')
                lb_shm_path = None
        else:
            global_logger.warning('/dev/shm is not supported by OS, creating ephemeral database in temp dir')

        fd, db_path = tempfile.mkstemp(dir=lb_shm_path, prefix='shedb-')

    if opts.verbosity_pinger:
        logging.get_logger('scheduler.worker_pinger').setLevel(opts.verbosity_pinger)
    try:
        asyncio.run(main_async(db_path, broadcast_interval=opts.broadcast_interval))
    except KeyboardInterrupt:
        global_logger.warning('SIGINT caught')
        global_logger.info('SIGINT caught. Scheduler is stopped now.')
    finally:
        if opts.ephemeral:
            assert fd is not None
            os.close(fd)
            os.unlink(db_path)


if __name__ == '__main__':
    main(sys.argv[1:])



from copy import copy
from lifeblood.basenode import BaseNode
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError

from typing import Iterable


def node_class():
    return FramerangeSplitter


class FramerangeSplitter(BaseNode):

    @classmethod
    def label(cls) -> str:
        return 'frame range splitter'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'frame', 'framerange', 'splitter', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'framerange_splitter'

    def __init__(self, name):
        super(FramerangeSplitter, self).__init__(name)
        with self.get_ui().initializing_interface_lock():
            self.get_ui().add_parameter('chunk size', 'chunk size in frames', NodeParameterType.INT, 10)

    def process_task(self, context):
        attrs = context.task_attributes()
        if 'frames' not in attrs:
            raise ProcessingError('attribute "frames" not found')
        frames = attrs['frames']
        if not isinstance(frames, list):
            return ProcessingResult()
        res = ProcessingResult()
        chunksize = context.param_value('chunk size')

        # res.set_attribute('frames', frames[:chunksize])
        split_into = 1 + (len(frames) - 1) // chunksize
        # yes we can split into 1 part. this should behave the same way as when splitting into multiple parts
        # in order to have consistent behaviour in the graph
        res.split_task(split_into)
        for i in range(split_into):
            res.set_split_task_attrib(i, 'frames', frames[chunksize*i:chunksize*(i+1)])
        return res

    def postprocess_task(self, context):
        return ProcessingResult()



import sys
import errno
import argparse
import asyncio
import signal
import time
import itertools
from types import MappingProxyType
import json
from .config import get_config
from .nethelpers import get_default_addr
from .broadcasting import await_broadcast
from .defaults import scheduler_port as default_scheduler_port, message_proxy_port
from .pulse_checker import PulseChecker
from .process_utils import create_worker_process, send_stop_signal_to_worker

from .logging import get_logger
# from .worker_pool_protocol import WorkerPoolProtocol
from .worker_pool_message_processor import WorkerPoolMessageProcessor
from .nethelpers import get_localhost
from .enums import WorkerState, WorkerType, ProcessPriorityAdjustment
from .defaults import worker_pool_port as default_worker_pool_port

from .net_messages.address import AddressChain, DirectAddress

from typing import Tuple, Dict, List, Optional


async def create_worker_pool(worker_type: WorkerType = WorkerType.STANDARD, *,
                             minimal_total_to_ensure=0, minimal_idle_to_ensure=0, maximum_total=256,
                             idle_timeout=10, worker_suspicious_lifetime=4, priority=ProcessPriorityAdjustment.NO_CHANGE, scheduler_address: Optional[AddressChain] = None):
    swp = WorkerPool(worker_type,
                     minimal_total_to_ensure=minimal_total_to_ensure, minimal_idle_to_ensure=minimal_idle_to_ensure, maximum_total=maximum_total,
                     idle_timeout=idle_timeout, worker_suspicious_lifetime=worker_suspicious_lifetime, priority=priority, scheduler_address=scheduler_address)
    await swp.start()
    return swp


class ProcData:
    __slots__ = {'process', 'id', 'state', 'state_entering_time', 'start_time', 'sent_term_signal'}

    def __init__(self, process: asyncio.subprocess.Process, id: int):
        self.process = process
        self.id = id
        self.state: WorkerState = WorkerState.OFF
        self.state_entering_time = 0
        self.start_time = time.time()
        self.sent_term_signal = False


class WorkerPool:  # TODO: split base class, make this just one of implementations
    def __init__(self, worker_type: WorkerType = WorkerType.STANDARD, *,
                 minimal_total_to_ensure=0, minimal_idle_to_ensure=0, maximum_total=256,
                 idle_timeout=10, worker_suspicious_lifetime=4, priority=ProcessPriorityAdjustment.NO_CHANGE,
                 scheduler_address: Optional[AddressChain] = None):
        """
        manages a pool of workers.
        :param worker_type: workers are created of given type
        :param minimal_total_to_ensure:  at minimum this amount of workers will be always upheld
        :param minimal_idle_to_ensure:  at minimum this amount of IDLE or OFF(as we assume they are OFF only while they are booting up) workers will be always upheld
        :param scheduler_address:  force created workers to use this scheduler address. otherwise workers will use their configuration
        """
        # local helper workers' pool
        self.__worker_pool: Dict[asyncio.Future, ProcData] = {}
        self.__workers_to_merge: List[ProcData] = []
        self.__pool_task = None
        # self.__worker_server: Optional[asyncio.AbstractServer] = None
        self.__message_proxy: Optional[WorkerPoolMessageProcessor] = None
        self.__stop_event = asyncio.Event()
        self.__server_closer_waiter = None
        self.__poke_event = asyncio.Event()
        self.__logger = get_logger(self.__class__.__name__.lower())
        self.__ensure_minimum_total = minimal_total_to_ensure
        self.__ensure_minimum_idle = minimal_idle_to_ensure
        self.__maximum_total = maximum_total
        self.__worker_type = worker_type
        self.__idle_timeout = idle_timeout  # after this amount of idling worker will be stopped if total count is above minimum
        self.__worker_priority = priority
        self.__scheduler_address = scheduler_address

        # workers are not created as singleshot, so lifetime of less then this should be considered a sign of possible error
        self.__suspiciously_short_process_time = worker_suspicious_lifetime

        self.__pulse_checker = None

        self.__id_to_procdata: Dict[int, ProcData] = {}
        self.__next_wid = 0
        # self.__my_addr = get_localhost()
        # self.__my_port = default_worker_pool_port()
        self.__message_proxy_address = None

        self.__poke_event.set()
        self.__stopped = False

    async def start(self):
        if self.__pool_task is not None and not self.__pool_task.done():
            return

        proxy_addr, proxy_port = get_default_addr(), message_proxy_port()
        for i in range(1024):  # somewhat big, but not too big
            self.__message_proxy = WorkerPoolMessageProcessor(self, (proxy_addr, proxy_port))  # TODO: config for other arguments
            try:
                await self.__message_proxy.start()
                break
            except OSError as e:
                if e.errno != errno.EADDRINUSE:
                    raise
                proxy_port += 1
                continue
        else:
            raise RuntimeError(f'could not find an opened port in range [{message_proxy_port()}-{proxy_port}]!')
        self.__message_proxy_address = DirectAddress.from_host_port(proxy_addr, proxy_port)

        self.__pool_task = asyncio.create_task(self.local_worker_pool_manager())

        self.__pulse_checker = PulseChecker(self.__scheduler_address, self.__message_proxy, interval=10, maximum_misses=10)
        self.__pulse_checker.add_pulse_fail_callback(self._on_pulse_fail)
        await self.__pulse_checker.start()
        self.__logger.debug(f'worker pool message protocol listening on {self.__message_proxy_address}')

    def stop(self):
        async def _server_closer():
            self.__pulse_checker.stop()
            await self.__pulse_checker
            await self.__pool_task  # ensure local manager is stopped before closing server. here it will ensure all workers are terminated
            self.__message_proxy.stop()
            await self.__message_proxy.wait_till_stops()
            self.__logger.info('message processor stopped')

        if self.__stopped:
            return
        self.__stop_event.set()  # stops local_worker_pool_manager
        self.__server_closer_waiter = asyncio.create_task(_server_closer())  # server will be closed here
        self.__stopped = True

    def __await__(self):
        return self.wait_till_stops().__await__()

    async def wait_till_stops(self):
        if self.__pool_task is None:
            return
        await self.__stop_event.wait()
        await self.__server_closer_waiter

    def is_stopping(self) -> bool:
        """
        True if worker pool is stopped or in process of stopping
        """
        return self.__stop_event.is_set()

    def is_pool_closed(self) -> bool:
        """
        True if main task has finished
        """
        return self.__pool_task.done()

    async def add_worker(self):
        if self.__stopped:
            self.__logger.warning('add_worker called after stop()')
            return
        if len(self.__id_to_procdata) + len(self.__workers_to_merge) >= self.__maximum_total:
            self.__logger.warning(f'maximum worker limit reached ({self.__maximum_total})')
            return
        args = [sys.executable, '-m', 'lifeblood.launch',
                '--loglevel', 'DEBUG',
                'worker',
                '--type', self.__worker_type.name,
                '--priority', self.__worker_priority.name,
                '--no-loop',
                '--id', str(self.__next_wid),
                '--pool-address', str(self.__message_proxy_address)]
        if self.__scheduler_address is not None:
            args += ['--scheduler-address',
                     AddressChain.join_address((
                         self.__message_proxy.listening_address(),
                         self.__scheduler_address
                     ))]

        self.__workers_to_merge.append(ProcData(await create_worker_process(args), self.__next_wid))
        self.__logger.debug(f'adding new worker (id: {self.__next_wid}) to the pool, total: {len(self.__workers_to_merge) + len(self.__worker_pool)}')
        self.__next_wid += 1
        self.__poke_event.set()

    def list_workers(self):
        return MappingProxyType(self.__id_to_procdata)

    def set_minimum_total_workers(self, minimum_total: int):
        self.__ensure_minimum_total = minimum_total

    def set_minimum_idle_workers(self, minimum_idle: int):
        self.__ensure_minimum_idle = minimum_idle

    def set_maximum_workers(self, maximum: int):
        self.__maximum_total = maximum

    #
    # local worker pool manager
    async def local_worker_pool_manager(self):
        """
        this task is responsible for local worker management.
        kill them if aborted
        :return:
        """
        async def _wait_and_reset_event(event: asyncio.Event, timeout):
            await asyncio.sleep(timeout)
            event.clear()

        check_timeout = 10
        stop_waiter = asyncio.create_task(self.__stop_event.wait())
        poke_waiter = asyncio.create_task(self.__poke_event.wait())
        no_adding_workers = asyncio.Event()
        wait_event_task = None
        try:
            while True:
                done, pending = await asyncio.wait(itertools.chain(self.__worker_pool.keys(), (stop_waiter, poke_waiter)), timeout=check_timeout, return_when=asyncio.FIRST_COMPLETED)
                time_to_stop = False
                if wait_event_task is not None and wait_event_task.done():
                    wait_event_task = None

                for x in done:
                    if x == stop_waiter:
                        time_to_stop = True
                        self.__logger.info('stopping worker pool...')
                        if not poke_waiter.done():
                            poke_waiter.cancel()
                        break
                    elif x == poke_waiter:
                        self.__poke_event.clear()
                        poke_waiter = asyncio.create_task(self.__poke_event.wait())
                        continue
                    # if not those 2 - x must be a process awaiting task
                    if (span := time.time() - self.__worker_pool[x].start_time) < self.__suspiciously_short_process_time:
                        self.__logger.warning(f'worker died within suspicious time threshold: {span}s. pausing worker creation for a bit')
                        if wait_event_task is None:
                            no_adding_workers.set()
                            wait_event_task = asyncio.create_task(_wait_and_reset_event(no_adding_workers, 5))
                    wid = self.__worker_pool[x].id
                    del self.__worker_pool[x]
                    del self.__id_to_procdata[wid]

                    self.__logger.debug(f'removing finished worker from the pool, total: {len(self.__workers_to_merge) + len(self.__worker_pool)}')
                if time_to_stop:
                    break
                for procdata in self.__workers_to_merge:
                    self.__worker_pool[asyncio.create_task(procdata.process.wait())] = procdata
                    self.__id_to_procdata[procdata.id] = procdata

                self.__workers_to_merge.clear()

                # check for idle workers
                idle_guys = len([k for k, v in self.__id_to_procdata.items() if v.state in (WorkerState.IDLE, WorkerState.OFF)])  # consider OFF ones as IDLEs that just boot up
                if idle_guys > self.__ensure_minimum_idle and len(self.__id_to_procdata) > self.__ensure_minimum_total:
                    max_to_kill = min(idle_guys - self.__ensure_minimum_idle, len(self.__id_to_procdata) - self.__ensure_minimum_total)
                    # if we above minimum - we can kill some idle ones
                    now = time.time()
                    for procdata in self.__worker_pool.values():
                        if max_to_kill <= 0:
                            break
                        if procdata.state != WorkerState.IDLE or now - procdata.state_entering_time < self.__idle_timeout or procdata.sent_term_signal:
                            continue
                        try:
                            send_stop_signal_to_worker(procdata.process)
                            procdata.sent_term_signal = True
                        except ProcessLookupError:
                            # probability is low, but this can happen. though if this happens often - something is wrong
                            self.__logger.warning("tried kill some idle workers, but it was already dead")
                        else:
                            max_to_kill -= 1
                            self.__poke_event.set()  # poke ourselves to clean up finished processes

                # ensure the ensure
                if not no_adding_workers.is_set():
                    just_added = 0
                    if len(self.__worker_pool) < self.__ensure_minimum_total:
                        for _ in range(self.__ensure_minimum_total - len(self.__worker_pool)):
                            await self.add_worker()
                            just_added += 1  # cuz add_worker will not add to __id_to_procdata or __worker_pool - we do on next iteration

                    if idle_guys + just_added < self.__ensure_minimum_idle:
                        for _ in range(self.__ensure_minimum_idle - idle_guys - just_added):
                            await self.add_worker()
                else:
                    self.__logger.debug('temporarily not adding workers')

            # debug logging
            self.__logger.debug(f'at pool closing, before cleanup: total workers: {len(self.__worker_pool)}, idle: {len([k for k, v in self.__id_to_procdata.items() if v.state in (WorkerState.IDLE, WorkerState.OFF)])}')
            # more verbose debug:
            if True:
                for wid, procdata in self.__id_to_procdata.items():
                    self.__logger.debug(f'worker id {wid}, pid {procdata.process.pid}: {procdata.state}')
        except asyncio.CancelledError:
            self.__logger.info('cancelled! stopping worker pool...')
            raise
        finally:
            async def _proc_waiter(proc: asyncio.subprocess.Process):
                try:
                    await asyncio.wait_for(proc.wait(), timeout=10)
                    self.__logger.debug(f'{proc.pid} has gracefully ended with {await proc.wait()}')
                except asyncio.TimeoutError:
                    self.__logger.warning('worker ignored SIGINT. killing instead.')
                    proc.kill()
                    await proc.wait()
                except Exception as e:
                    self.__logger.exception('very unexpected exception. pretending like it hasn\'t happened')

            # cleanup
            wait_tasks = []
            for procdata in itertools.chain(self.__worker_pool.values(), self.__workers_to_merge):
                try:
                    self.__logger.debug(f'sending SIGTERM to {procdata.process.pid}')
                    send_stop_signal_to_worker(procdata.process)
                except ProcessLookupError:
                    continue
                wait_tasks.append(_proc_waiter(procdata.process))
            await asyncio.gather(*wait_tasks)
            await asyncio.gather(*self.__worker_pool.keys())  # since all processes are killed now - this SHOULD take no time at all

            self.__logger.info('worker pool stopped')
        # tidyup
        for fut in self.__worker_pool:
            if not fut.done():
                fut.cancel()

    #
    # callbacks
    async def _on_pulse_fail(self):
        self.stop()

    #
    # callbacks from protocol
    async def _worker_state_change(self, worker_id: int, state: WorkerState):
        if worker_id not in self.__id_to_procdata:
            self.__logger.warning(f'reported state {state} for worker {worker_id} that DOESN\'T BELONG TO US')
            return

        if self.__id_to_procdata[worker_id].state != state:
            self.__id_to_procdata[worker_id].state = state
            self.__id_to_procdata[worker_id].state_entering_time = time.time()
            self.__poke_event.set()


async def async_main(argv):
    logger = get_logger('simple_worker_pool')
    parser = argparse.ArgumentParser('lifeblood pool simple')
    parser.add_argument('--min-idle', '-m',
                        dest='minimal_idle_to_ensure',
                        default=1, type=int,
                        help='worker pool will ensure at least this amount of workers is up idle (default=1)')
    parser.add_argument('--min-total',
                        dest='minimal_total_to_ensure',
                        default=0, type=int,
                        help='worker pool will ensure at least this amount of workers is up total (default=0)')
    parser.add_argument('--max', '-M',
                        dest='maximum_total',
                        default=256, type=int,
                        help='no more than this amount of workers will be run locally at the same time (default=256)')
    parser.add_argument('--priority', choices=tuple(x.name for x in ProcessPriorityAdjustment), default=ProcessPriorityAdjustment.LOWER.name, help='pass to spawned workers: adjust child process priority')

    opts = parser.parse_args(argv)
    opts.priority = [x for x in ProcessPriorityAdjustment if x.name == opts.priority][0]  # there MUST be exactly 1 match

    graceful_closer_no_reentry = False

    def graceful_closer(*args):
        nonlocal graceful_closer_no_reentry
        if graceful_closer_no_reentry:
            print('DOUBLE SIGNAL CAUGHT: ALREADY EXITING')
            return
        graceful_closer_no_reentry = True
        logger.info('SIGINT/SIGTERM caught')
        nonlocal noloop
        noloop = True
        stop_event.set()
        if pool:
            pool.stop()

    noasync_do_close = False

    def noasync_windows_graceful_closer_event(*args):
        nonlocal noasync_do_close
        noasync_do_close = True

    async def windows_graceful_closer():
        while not noasync_do_close:
            await asyncio.sleep(1)
        graceful_closer()

    logger.debug(f'starting {__name__} with: ' + ', '.join(f'{key}={val}' for key, val in opts.__dict__.items()))
    pool = None
    noloop = False  # TODO: add arg

    # override event handlers
    win_signal_waiting_task = None
    try:
        asyncio.get_event_loop().add_signal_handler(signal.SIGINT, graceful_closer)
        asyncio.get_event_loop().add_signal_handler(signal.SIGTERM, graceful_closer)
    except NotImplementedError:  # solution for windows
        signal.signal(signal.SIGINT, noasync_windows_graceful_closer_event)
        signal.signal(signal.SIGBREAK, noasync_windows_graceful_closer_event)
        win_signal_waiting_task = asyncio.create_task(windows_graceful_closer())
    #

    stop_event = asyncio.Event()
    stop_task = asyncio.create_task(stop_event.wait())
    config = get_config('worker')

    start_attempt_cooldown = 0
    while True:
        if await config.get_option('worker.listen_to_broadcast', True):
            logger.info('listening for scheduler broadcasts...')
            broadcast_task = asyncio.create_task(await_broadcast('lifeblood_scheduler'))
            done, _ = await asyncio.wait((broadcast_task, stop_task), return_when=asyncio.FIRST_COMPLETED)
            if stop_task in done:
                broadcast_task.cancel()
                logger.info('broadcast listening cancelled')
                break
            assert broadcast_task.done()
            message = await broadcast_task
            scheduler_info = json.loads(message)
            logger.debug('received', scheduler_info)
            addr = AddressChain(scheduler_info['message_address'])
        else:
            if stop_event.is_set():
                break
            logger.info('boradcast listening disabled')
            start_attempt_cooldown = 10
            if not config.has_option_noasync('worker.scheduler_address'):
                raise RuntimeError('worker.scheduler_address config option must be provided')
            addr = AddressChain(await config.get_option('worker.scheduler_address', None))
            logger.debug(f'using {addr}')

        try:
            pool = await create_worker_pool(WorkerType.STANDARD, scheduler_address=addr, **opts.__dict__)
        except Exception:
            logger.exception('could not start the pool')
            await asyncio.sleep(start_attempt_cooldown)
        else:
            await pool.wait_till_stops()
            logger.info('pool quited')
        if noloop:
            break

    if win_signal_waiting_task is not None:
        if not win_signal_waiting_task.done():
            win_signal_waiting_task.cancel()
    logger.info('pool loop stopped')


def main(argv):
    try:
        asyncio.run(async_main(argv))
    except KeyboardInterrupt:
        get_logger('simple_worker_pool').warning('SIGINT caught where it wasn\'t supposed to be caught')

